# Copyright 2025 The JAX Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Layout and transform inference pass for the MLIR Mosaic GPU dialect."""

# mypy has been causing more problems than it solves here. Disable it for these
# files. We have pytype checks anyway.
# mypy: ignore-errors

from __future__ import annotations

from collections.abc import Callable, Iterator, Sequence, Set
import dataclasses
import enum
from functools import partial
import itertools
import re
from typing import Any, assert_never, cast

from jax._src.lib import mosaic_gpu_dialect as mgpu  # noqa: F401
from jax._src.lib.mlir import ir
from jax._src.lib.mlir.dialects import arith
from jax._src.lib.mlir.dialects import gpu
from jax._src.lib.mlir.dialects import math as mlir_math
from jax._src.lib.mlir.dialects import memref
from jax._src.lib.mlir.dialects import scf
from jax._src.lib.mlir.dialects import vector

from . import equations as eqns
from . import fragmented_array as fa
from . import inference_utils
from . import launch_context as lc
from . import layouts as layouts_lib
from . import tcgen05
from . import utils
import numpy as np


class VariableType(enum.IntEnum):
  """The type of a variable.

  Variables here are either operands or results of MLIR operations.
  """
  OPERAND = 0
  RESULT = 1


class MemorySpace(enum.Enum):
  """The memory space of a variable."""
  REG = enum.auto()
  SMEM = enum.auto()
  TMEM = enum.auto()


_op_name_regex = re.compile(r"^(%\d+ = )?\S+")

@dataclasses.dataclass(frozen=True)
class OperandOrResult:
  """A unique identifier for a variable."""
  # A MLIR operation.
  operation: ir.OpView
  # Whether this represents an operand or a result.
  type: VariableType
  # The index of the operand/result within the op's operands/results.
  index: int

  @property
  def value(self) -> ir.Value:
    """Returns the IR value corresponding to this operand or result."""
    if self.type == VariableType.OPERAND:
      return self.operation.operands[self.index]
    else:
      return self.operation.results[self.index]

  @property
  def memory_space(self) -> MemorySpace:
    """Returns the memory space associated with this operand or result."""
    type = self.value.type
    if ir.VectorType.isinstance(type):
      return MemorySpace.REG
    assert ir.MemRefType.isinstance(type)
    if utils.is_tmem_ref(type):
      return MemorySpace.TMEM
    elif utils.is_smem_ref(type):
      return MemorySpace.SMEM
    raise ValueError(f"Unsupported memory space for: {type}")

  def __str__(self):
    match = _op_name_regex.match(str(self.operation))
    assert match is not None
    if self.type == VariableType.OPERAND:
      return f"{match.group(0)}:o-{self.index}"
    else:
      return f"{match.group(0)}:r-{self.index}"


@dataclasses.dataclass(frozen=True)
class Hint:
  """Hints are used to model propagation of layouts across operations.

  Since using `relayout`s is always an option in principle, propagation across
  ops can not rely only on an equation system. Instead, we introduce hints as
  an equation-like form of "soft constraints", i.e., it suggests that
  `variable` should be equal to `expression`.
  """
  variable: eqns.Variable
  expression: eqns.Expression

  def __str__(self):
    return f"{self.variable} ?= {self.expression}"


def extract_constant_from_least_replicated_expression_for_hint(
    expressions: tuple[eqns.Expression, ...],
) -> eqns.Constant | None:
  choices: list[eqns.Constant] = []
  for e in expressions:
    if (red := extract_constant_for_hint(e)) is not None:
      choices.append(red)

  if not choices:
    return None

  # We reduce the expression here in order to recover an unambiguous least
  # replicated layout if it exists.
  maybe_choice = eqns.reduce_expression(
      eqns.LeastReplicated(tuple(choices)), {}
  )

  if isinstance(maybe_choice, eqns.Unsatisfiable):
    # TODO(bchetioui): consider other choices.
    return choices[0]

  assert isinstance(maybe_choice, eqns.Constant)
  return maybe_choice


def extract_constant_from_most_replicated_expression_for_hint(
    expressions: tuple[eqns.Expression, ...],
) -> eqns.Constant | None:
  assert len(expressions) >= 1
  choices: list[eqns.Constant] = []
  for e in expressions:
    if (red := extract_constant_for_hint(e)) is not None:
      choices.append(red)

  if not choices:
    return None

  maybe_choice = eqns.reduce_expression(
      eqns.MostReplicated(tuple(choices)), {}
  )

  if isinstance(maybe_choice, eqns.Unsatisfiable):
    # TODO(bchetioui): consider other choices.
    return choices[0]

  assert isinstance(maybe_choice, eqns.Constant)
  return maybe_choice


def extract_constant_from_broadcast_in_dim_expression_for_hint(
    e: eqns.BroadcastInDim,
) -> eqns.RegisterLayout | None:
  if not isinstance(e.expression, eqns.RegisterLayout):
    return None

  reduced_layout = e.expression.value

  wgmma_tm, wgmma_tn = fa.WGMMA_LAYOUT.base_tile_shape
  # TODO(bchetioui): enable generators to handle TCGEN05 layout from WGMMA_COL.
  if reduced_layout == fa.WGMMA_COL_LAYOUT and e.axes == (1,) and e.shape[0] % wgmma_tm == 0:
    return eqns.RegisterLayout(fa.WGMMA_LAYOUT)

  if reduced_layout == fa.WGMMA_ROW_LAYOUT and e.axes == (0,) and e.shape[1] % wgmma_tn == 0:
    return eqns.RegisterLayout(fa.WGMMA_LAYOUT)

  tcgen05_tm, _ = fa.TCGEN05_LAYOUT.base_tile_shape
  if reduced_layout == fa.TCGEN05_ROW_LAYOUT and e.axes == (0,) and e.shape[0] % tcgen05_tm == 0:
    return eqns.RegisterLayout(fa.TCGEN05_LAYOUT)

  return None


def extract_constant_for_hint(e: eqns.Expression) -> eqns.Constant | None:
  """Attempts to extract a `ConstantExpression` from a `Hint`'s `Expression`.

  Returns `None` if no `ConstantExpression` could be reasonably extracted.
  """
  match e:
    case eqns.Constant():
      return e
    case eqns.LeastReplicated():
      return extract_constant_from_least_replicated_expression_for_hint(e.expressions)
    case eqns.MostReplicated():
      return extract_constant_from_most_replicated_expression_for_hint(e.expressions)
    case eqns.BroadcastInDim():
      return extract_constant_from_broadcast_in_dim_expression_for_hint(e)
    case eqns.Variable():
      return None
    case _:
      raise NotImplementedError(f"Unsupported expression type: {type(e)}")


def extract_variable_assignment_from_hint(
    hint: Hint,
) -> tuple[eqns.Variable, eqns.Constant] | None:
  """Attempts to extract a single variable assignment from a `Hint`."""
  # TODO(bchetioui): make this a generator. This will allow us to maybe extract
  # different assignments that satisfy a replication constraint in the case
  # where replicated expressions are incompatible and several extractions are
  # possible.
  red = extract_constant_for_hint(hint.expression)
  return (hint.variable, red) if red is not None else None


def reduce_hints(
    hints: Sequence[Hint], assignments: dict[eqns.Variable, eqns.Constant]
) -> list[Hint]:
  """Reduces a sequence of `Hint`s.

  We reduce the `Hint`s' expressions, drop `Unsatisfiable` hints, and drop
  `Hint`s pertaining to pre-existing assignments.
  """
  new_hints: list[Hint] = []
  for h in hints:
    if h.variable not in assignments:
      reduced_expression = eqns.reduce_expression(h.expression, assignments)
      if isinstance(reduced_expression, eqns.Unsatisfiable):
        continue
      new_hints.append(dataclasses.replace(h, expression=reduced_expression))

  return new_hints


def _strided_layout_for_variable(
    variable: eqns.Variable,
) -> fa.WGStridedFragLayout | None:
  """Returns a strided layout for the given variable.

  If the given variable cannot have a strided layout, returns `None`.
  """
  # TODO(bchetioui): should we make variables carry a shape as well, to make
  # things easier?
  type = variable.key.value.type
  assert ir.VectorType.isinstance(type)
  return fa.WGStridedFragLayout.from_shaped_type(type)


def _extract_tiling_candidate(
    divides: eqns.Divides, num_tiled_dims: int
) -> Iterator[tuple[eqns.Variable, eqns.Constant]]:
  if not isinstance(divides.expr, eqns.Variable):
    return
  if num_tiled_dims > len(divides.dimensions_to_tile):
    # TODO(b/447079781): Support this case, by just assuming 0 (no constraints).
    return

  if num_tiled_dims == 0:
    yield divides.expr, eqns.SMEMTiling(None)
    return

  static_tiling = []
  dynamic_tiling = []
  for dim in divides.dimensions_to_tile[-num_tiled_dims:]:
    assert isinstance(dim[0], int)
    static_tiling.append(dim[0])
    # Any remaining values are dynamic. Use 1 for those to be safe.
    dynamic_tiling.append(dim[0] if len(dim) == 1 else 1)

  # The static tiling ignores dynamic values. Try this first as it yields
  # larger tiles that are likely to have better performance.
  yield divides.expr, eqns.SMEMTiling(lc.TileTransform(tuple(static_tiling)))

  if dynamic_tiling != static_tiling:
    # If that does not work, we could use a smaller, safe tiles by yielding
    # dynamic_tiling here. However, performance will be worse, and we choose to
    # return instead.
    return

def _extract_layout_candidates_from_memory_space_transfers(
    constraint: eqns.IsTransferable,
    divides_per_var: dict[eqns.Variable, list[eqns.Divides]],
) -> Iterator[tuple[eqns.Variable, eqns.Constant]]:
  """Attempts to extract variable assignments from a `Constraint`."""
  # This code assumes that the `IsTransferable` constraint is bidirectional.
  # This is currently true for TMEM <-> REG transfers and SMEM <-> REG
  # transfers.
  src, tgt = constraint.source, constraint.target
  match src, tgt:
    case eqns.Variable(), eqns.Constant():
      variable, constant = src, tgt
    case eqns.Constant(), eqns.Variable():
      variable, constant = tgt, src
    case _:
      return

  assert isinstance(variable, eqns.Variable)  # Satisfy type checkers.
  if isinstance(constant, eqns.RegisterLayout):
    layout = constant.value
    if variable.key.memory_space == MemorySpace.TMEM:
      for packing in (1, 2, 4, 8):
        for tmem_layout, reg_layout in constraint.supported_tmem_transfers(
            packing
        ):
          if layout == reg_layout:
            yield variable, eqns.TMEMLayout(tmem_layout)
    elif variable.key.memory_space == MemorySpace.SMEM:
      if inference_utils.is_mma_layout(layout):
        tiling = _infer_tiling_for_mma_ref(
            variable.key.value.type,
            max_swizzle=mgpu.SwizzlingMode.k128ByteSwizzle
        )
        assert len(tiling) == 2
        divides = divides_per_var.get(variable, [])
        divides.append(eqns.Divides(variable, ((tiling[0],), (tiling[1],))))
        [divides] = eqns.merge_divides_constraints(divides)
        assert isinstance(divides, eqns.Divides)
        yield from _extract_tiling_candidate(divides, len(tiling))
      else:
        # An empty tiling is valid here but we don't yield it in order to
        # avoid duplicating the empty tiling yielded by the caller.
        return

  elif isinstance(constant, eqns.TMEMLayout):
    layout = constant.value
    packing = layout.vector_length
    for tmem_layout, reg_layout in constraint.supported_tmem_transfers(packing):
      if layout == tmem_layout:
        yield variable, eqns.RegisterLayout(reg_layout)


def _divides_per_var(
    constraints: Sequence[eqns.Constraint],
) -> dict[eqns.Variable, list[eqns.Divides]]:
  """Returns all Divides constraints per variable."""
  result: dict[eqns.Variable, list[eqns.Divides]] = {}
  for constraint in constraints:
    match constraint:
      case eqns.Divides(expr=expr) if isinstance(expr, eqns.Variable):
        result.setdefault(expr, []).append(constraint)
  return result


def _extract_variable_assignments_from_constraints(
    constraints: Sequence[eqns.Constraint],
) -> Iterator[tuple[eqns.Variable, eqns.Constant]]:
  """Attempts to extract variable assignments from all constraints."""
  dpv = _divides_per_var(constraints)
  for c in constraints:
    match c:
      case eqns.IsTransferable():
        yield from _extract_layout_candidates_from_memory_space_transfers(c, dpv)


def conjure_assignment(
    unknowns: Set[eqns.Variable],
    equation_system: eqns.EquationSystem,
    hints: Sequence[Hint],
) -> Iterator[tuple[eqns.Variable, eqns.Constant]]:
  """Attempts to conjure an assignment for an unknown variable."""
  # TODO(allanrenucci): We should be able to short-circuit the search here if
  # the constraint is not satisfiable.
  yield from _extract_variable_assignments_from_constraints(
      equation_system.constraints
  )

  for hint in hints:
    if (assignment := extract_variable_assignment_from_hint(hint)) is not None:
      yield assignment

  # Here, we have not managed to find an assignment for all the unknown
  # variables, and our hints have not proven sufficient to unblock us. We now
  # try to introduce new arbitrary (valid) assignments into the system, and
  # hope that they turn out to be compatible with the equation system.
  for variable in unknowns:
    if variable in equation_system.assignments:
      continue
    # Try to instantiate a single variable to a strided layout and see if it
    # reduces the system.
    if variable.key.memory_space == MemorySpace.REG:
      layout = _strided_layout_for_variable(variable)
      if layout is not None:
        yield variable, eqns.RegisterLayout(layout)
    elif variable.key.memory_space == MemorySpace.SMEM:
      yield variable, eqns.SMEMTiling(None)


def find_assignments_for(
    unknowns: Set[eqns.Variable],
    equation_system: eqns.EquationSystem,
    hints: Sequence[Hint],
) -> dict[eqns.Variable, eqns.Constant] | eqns.Unsatisfiable:
  """Attempts to find assignments that satisfy `equation_system` for `unknowns`.

  Args:
    unknowns: the set of variables that are unknown.
    equation_system: the equation system to satisfy.
    hints: a list of hints that may be used to introduce new assignments.

  Returns:
    - Unsatisfiable() if the equation system has unsatisfiable constraints.
    - A dictionary assigning all the unknown variables to `ConstantExpression`s
      such that the assignment satisfies the equation system otherwise.
  """
  equation_system = eqns.reduce(equation_system)
  if isinstance(equation_system, eqns.Unsatisfiable):
    return eqns.Unsatisfiable()

  remaining_unknowns = unknowns - equation_system.assignments.keys()
  # In this case, we have determined an assignment for all the unknown
  # variables. Return their respective assignment.
  if not remaining_unknowns:
    return {v: k for v, k in equation_system.assignments.items() if v in unknowns}

  # Reduce the expressions in the remaining hints based on the current
  # assignments, and eliminate hints that pertain to variables that already
  # have an assignment.
  hints = reduce_hints(hints, equation_system.assignments)

  # If unknowns remain and we have fully reduced the system, we may still
  # be able to make progress by extracting an assignment from a `Hint`. This
  # new assignment could make the system unsatisfiable, so we use a recursive
  # call to be able to backtrack if necessary.
  for assignment in conjure_assignment(remaining_unknowns, equation_system, hints):
    variable, expr = assignment
    new_equation_system = (
        eqns.EquationSystem(assignments={variable: expr}) & equation_system)
    if isinstance(new_equation_system, eqns.Unsatisfiable):
      # This assignment is not compatible with the equation system.
      continue
    solution = find_assignments_for(unknowns, new_equation_system, hints)
    if isinstance(solution, eqns.Unsatisfiable):
      # This assignment is not compatible with the equation system.
      continue
    return solution

  # TODO(bchetioui): should we have a way to give a useful dump to the user
  # here, perhaps indicating what to layout cast.
  return eqns.Unsatisfiable()


@dataclasses.dataclass()
class DerivationContext:
  """Holds context information used for deriving an equation system."""
  # TODO(b/447079781): Remove once SMEM transform inference is implemented.
  enable_smem_inference : bool = False
  # A map of `OperandOrResult` to the variable that it is associated with.
  variable_for_operand_or_result: dict[OperandOrResult, eqns.Variable] = (
      dataclasses.field(default_factory=dict, init=False)
  )
  # A map of `eqns.Variable` to all the `OperandOrResult`s that it is associated
  # with.
  operand_and_results_for_variable: OperandOrResultsForVariable = (
      dataclasses.field(default_factory=dict, init=False)
  )

  def update(self, mapping: OperandOrResultsForVariable) -> None:
    for variable, operand_and_results in mapping.items():
      if variable in self.operand_and_results_for_variable:
        self.operand_and_results_for_variable[variable].extend(operand_and_results)
      else:
        self.operand_and_results_for_variable[variable] = operand_and_results
      for operand_or_result in operand_and_results:
        assert operand_or_result not in self.variable_for_operand_or_result
        self.variable_for_operand_or_result[operand_or_result] = variable

  def producer_ref(self, operand: OperandOrResult) -> eqns.Variable:
    """Returns the producer reference variable for the given operand."""
    return self.variable_for_operand_or_result[producer_result(operand)]


OperandOrResultsForVariable = dict[eqns.Variable, list[OperandOrResult]]

# An equation system derivation rule is a function that takes an MLIR operation
# and returns an equation system, a mapping from variables to operand/result
# identifiers, and a list of hints.
#
# The intended meaning of the mapping is that, for each identifier in the list
# keyed by a given variable, the MLIR operand/result corresponding to that
# identifier has the same layout as the variable.
#
# An `EquationSystemDerivationRule` must return a mapping such that the
# identifier corresponding to each operand/result must appear in the mapping,
# and each identifier in the mapping must be keyed by exactly one variable.
# Lastly, the mapping must only refer to variables and operands/results that
# correspond to the given operation.
EquationSystemDerivationRule = Callable[
    [DerivationContext, ir.OpView],
    tuple[eqns.EquationSystem, OperandOrResultsForVariable, list[Hint]],
]
_equation_system_derivation_rules: dict[str, EquationSystemDerivationRule] = {}


def _add_equation_system_derivation_rule(op: type[ir.OpView]):
  def wrapper(rule: EquationSystemDerivationRule):
    if op is not None:
      _equation_system_derivation_rules[op.OPERATION_NAME] = rule  # pytype: disable=attribute-error
    return rule
  return wrapper


def is_vector(v: ir.Value) -> bool:
  return ir.VectorType.isinstance(v.type)


def _is_smem_ref(v: ir.Value) -> bool:
  return ir.MemRefType.isinstance(v.type) and utils.is_smem_ref(v)


def _is_tmem_ref(v: ir.Value) -> bool:
  return ir.MemRefType.isinstance(v.type) and utils.is_tmem_ref(v)


def _pointwise_op_equation_system(
    ctx: DerivationContext,
    op: ir.OpView,
) -> tuple[eqns.EquationSystem, OperandOrResultsForVariable, list[Hint]]:
  del ctx
  all_operands_and_results = vector_operands_and_results(op)
  variable = eqns.Variable(all_operands_and_results[0])
  return eqns.EquationSystem(), {variable: all_operands_and_results}, []


for op in [
    arith.AddIOp,
    arith.AddFOp,
    arith.AndIOp,
    arith.BitcastOp,
    arith.CmpFOp,
    arith.CmpIOp,
    arith.ExtFOp,
    arith.ExtSIOp,
    arith.ExtUIOp,
    arith.FPToSIOp,
    arith.FPToUIOp,
    arith.MaximumFOp,
    arith.MaxUIOp,
    arith.MaxSIOp,
    arith.MinimumFOp,
    arith.MinUIOp,
    arith.MinSIOp,
    arith.MulIOp,
    arith.MulFOp,
    arith.OrIOp,
    arith.FloorDivSIOp,
    arith.DivUIOp,
    arith.DivFOp,
    arith.RemUIOp,
    arith.RemSIOp,
    arith.RemFOp,
    arith.SIToFPOp,
    arith.UIToFPOp,
    arith.SubIOp,
    arith.SubFOp,
    arith.TruncFOp,
    arith.TruncIOp,
    arith.XOrIOp,
    mlir_math.ExpOp,
    mlir_math.Exp2Op,
    mlir_math.LogOp,
    mlir_math.RsqrtOp,
    mlir_math.TanhOp,
]:
  _add_equation_system_derivation_rule(op)(_pointwise_op_equation_system)


def _divides_constraint_from_indices(
    var: eqns.Variable,
    indices: Sequence[ir.Value],
) -> eqns.Divides:
  """Returns a Divides constraint from the given load/store indices."""
  dimensions_to_tile : list[tuple[int | ir.Value, ...]] = []
  for i in indices:
    index_defining_op = i.owner.opview
    if isinstance(index_defining_op, arith.ConstantOp):
      dimensions_to_tile.append((index_defining_op.literal_value,))
    else:
      dimensions_to_tile.append((i,))
  return eqns.Divides(var, tuple(dimensions_to_tile))


@_add_equation_system_derivation_rule(vector.LoadOp)
def _vector_load_equation_system(
    ctx: DerivationContext,
    op: vector.LoadOp,
) -> tuple[eqns.EquationSystem, OperandOrResultsForVariable, list[Hint]]:
  # TODO(b/447079781): Investigate whether we should check for contiguous
  # strides here. An initial implementaiton of this failed the
  # test_gmem_to_smem_with_multiple_smem_indexers_and_transforms test, but
  # we should confirm that this is properly supported.

  # Registers
  dest = OperandOrResult(op, VariableType.RESULT, 0)
  dest_var = eqns.Variable(dest)
  operand_or_results_for_variable = {dest_var: [dest]}
  constraints = [
      eqns.Distinct(
          dest_var,
          eqns.RegisterLayout(
              fa.WGSplatFragLayout(shape=tuple(ir.ShapedType(op.result.type).shape))
          ),
      ),
  ]

  # SMEM
  if ctx.enable_smem_inference and utils.is_smem_ref(op.base):
    source = OperandOrResult(op, VariableType.OPERAND, 0)
    source_var = ctx.producer_ref(source)
    operand_or_results_for_variable[source_var] = [source]
    constraints.extend([
        eqns.IsTransferable(
            source=source_var,
            target=dest_var,
            shape=tuple(ir.ShapedType(op.result.type).shape),
        ),
        _divides_constraint_from_indices(source_var, op.indices),
    ])

  system = eqns.EquationSystem(constraints=constraints)
  return system, operand_or_results_for_variable, []


@_add_equation_system_derivation_rule(vector.StoreOp)
def _vector_store_equation_system(
    ctx: DerivationContext,
    op: vector.StoreOp,
) -> tuple[eqns.EquationSystem, OperandOrResultsForVariable, list[Hint]]:
  # TODO(b/447079781): Investigate whether we should check for contiguous
  # strides here. An initial implementaiton of this failed the
  # test_gmem_to_smem_with_multiple_smem_indexers_and_transforms test, but
  # we should confirm that this is properly supported.

  # Registers
  value = OperandOrResult(op, VariableType.OPERAND, 0)
  value_var = eqns.Variable(value)
  operand_or_results_for_variable = {value_var: [value]}

  # SMEM
  constraints = []
  if ctx.enable_smem_inference and utils.is_smem_ref(op.base):
    dest = OperandOrResult(op, VariableType.OPERAND, 1)
    dest_var = ctx.producer_ref(dest)
    operand_or_results_for_variable[dest_var] = [dest]
    constraints = [
        eqns.IsTransferable(
            source=value_var,
            target=dest_var,
            shape=tuple(ir.ShapedType(op.base.type).shape),
        ),
        _divides_constraint_from_indices(dest_var, op.indices),
    ]

  system = eqns.EquationSystem(constraints=constraints)
  return system, operand_or_results_for_variable, []


@_add_equation_system_derivation_rule(mgpu.OptimizationBarrierOp)
def _optimization_barrier_equation_system(
    ctx: DerivationContext,
    op: ir.OpView,
) -> tuple[eqns.EquationSystem, OperandOrResultsForVariable, list[Hint]]:
  del ctx
  operand_or_results_for_variable: OperandOrResultsForVariable = {}

  for i, operand in enumerate(op.operands):
    if not is_vector(operand):
      continue
    variable = eqns.Variable(OperandOrResult(op, VariableType.OPERAND, i))
    operand_or_results_for_variable[variable] = [
        OperandOrResult(op, VariableType.OPERAND, i),
        OperandOrResult(op, VariableType.RESULT, i)
    ]

  return eqns.EquationSystem(), operand_or_results_for_variable, []


@_add_equation_system_derivation_rule(vector.BroadcastOp)
def _vector_splat_equation_system(
    ctx: DerivationContext,
    op: ir.OpView,
) -> tuple[eqns.EquationSystem, OperandOrResultsForVariable, list[Hint]]:
  del ctx
  result = OperandOrResult(op, VariableType.RESULT, 0)
  variable = eqns.Variable(result)
  layout = fa.WGSplatFragLayout(tuple(cast(ir.ShapedType, op.result.type).shape))
  system = eqns.EquationSystem(
      assignments={variable: eqns.RegisterLayout(layout)}
  )
  return system, {variable: [result]}, []


@_add_equation_system_derivation_rule(arith.ConstantOp)
def _constant_equation_system(
    ctx: DerivationContext,
    constant_op: arith.ConstantOp,
) -> tuple[eqns.EquationSystem, OperandOrResultsForVariable, list[Hint]]:
  del ctx
  value = constant_op.value
  result = OperandOrResult(constant_op, VariableType.RESULT, 0)
  variable = eqns.Variable(result)
  shape = tuple(ir.ShapedType(constant_op.result.type).shape)
  if (
      ir.DenseElementsAttr.isinstance(value)
      and ir.DenseElementsAttr(value).is_splat
  ):
    layout = fa.WGSplatFragLayout(shape=shape)
    system = eqns.EquationSystem(
        assignments={variable: eqns.RegisterLayout(layout)}
    )
  else:
    constant_is_not_splat = eqns.Distinct(
        variable,
        eqns.RegisterLayout(fa.WGSplatFragLayout(shape=shape)),
    )
    system = eqns.EquationSystem(constraints=[constant_is_not_splat])

  return system, {variable: [result]}, []


def _terminator(
    block: ir.Block, expected_terminator: type[ir.OpView]
) -> ir.OpView:
  """Returns the terminator of the given block.

  Checks that the terminator is of the expected type.
  """
  terminator = block.operations[len(block.operations) - 1]
  assert isinstance(terminator, expected_terminator)
  return terminator.opview


@_add_equation_system_derivation_rule(scf.ForOp)
def _for_equation_system(
    ctx: DerivationContext,
    op: scf.ForOp,
) -> tuple[eqns.EquationSystem, OperandOrResultsForVariable, list[Hint]]:
  [block] = op.region.blocks
  yield_op = _terminator(block, scf.YieldOp)
  operand_or_results_for_variable: OperandOrResultsForVariable = {}

  # Account for the lower bound, upper bound, and step of the loop, which appear
  # in the operands but not in the results.
  num_leading_args = 3
  for index, o in enumerate(op.operands):
    if not is_vector(o) and not (ctx.enable_smem_inference and _is_smem_ref(o)):
      continue
    result_index = index - num_leading_args
    operand = OperandOrResult(op, VariableType.OPERAND, index)
    result = OperandOrResult(op, VariableType.RESULT, result_index)
    yield_operand = OperandOrResult(
        yield_op, VariableType.OPERAND, result_index
    )
    var = eqns.Variable(operand) if is_vector(o) else ctx.producer_ref(operand)
    operand_or_results_for_variable[var] = [operand, result, yield_operand]

  return eqns.EquationSystem(), operand_or_results_for_variable, []


@_add_equation_system_derivation_rule(scf.WhileOp)
def _while_equation_system(
    ctx: DerivationContext,
    op: scf.WhileOp,
) -> tuple[eqns.EquationSystem, OperandOrResultsForVariable, list[Hint]]:
  del ctx
  [before_block] = op.before.blocks
  [after_block] = op.after.blocks
  cond_op = _terminator(before_block, scf.ConditionOp)
  yield_op = _terminator(after_block, scf.YieldOp)

  operand_or_results_for_variable: OperandOrResultsForVariable = {}

  for operand_or_result in vector_operands_and_results(op):
    match operand_or_result.type:
      case VariableType.OPERAND:
        yield_operand = OperandOrResult(
            yield_op, VariableType.OPERAND, operand_or_result.index
        )
        operand_or_results_for_variable[eqns.Variable(operand_or_result)] = [
            operand_or_result,
            yield_operand,
        ]
      case VariableType.RESULT:
        # Increment by 1 to account for the conditional.
        cond_operand = OperandOrResult(
            cond_op, VariableType.OPERAND, operand_or_result.index + 1
        )
        operand_or_results_for_variable[eqns.Variable(operand_or_result)] = [
            operand_or_result,
            cond_operand,
        ]
      case _ as never:
        assert_never(never)

  return eqns.EquationSystem(), operand_or_results_for_variable, []


@_add_equation_system_derivation_rule(scf.IndexSwitchOp)
def _index_switch_equation_system(
    ctx: DerivationContext,
    op: scf.IndexSwitchOp,
) -> tuple[eqns.EquationSystem, OperandOrResultsForVariable, list[Hint]]:
  del ctx
  operand_or_results_for_variable: OperandOrResultsForVariable = {
      eqns.Variable(o): [o] for o in vector_operands_and_results(op)
  }
  for region in op.regions:
    [block] = region.blocks
    yield_op = _terminator(block, scf.YieldOp)
    for operand_or_result in operand_or_results_for_variable.keys():
      assert operand_or_result.key.type == VariableType.RESULT
      yield_operand = OperandOrResult(
          yield_op, VariableType.OPERAND, operand_or_result.key.index
      )
      operand_or_results_for_variable[operand_or_result].append(yield_operand)

  return eqns.EquationSystem(), operand_or_results_for_variable, []


@_add_equation_system_derivation_rule(mgpu.LayoutCastOp)
def _layout_cast_equation_system(
    ctx: DerivationContext,
    op: mgpu.LayoutCastOp,
) -> tuple[eqns.EquationSystem, OperandOrResultsForVariable, list[Hint]]:
  del ctx
  operand = OperandOrResult(op, VariableType.OPERAND, 0)
  result = OperandOrResult(op, VariableType.RESULT, 0)
  variable = eqns.Variable(operand)
  out_layout = eqns.RegisterLayout(layouts_lib.from_layout_attr(op.new_layout))
  return (
      eqns.EquationSystem(assignments={variable: out_layout}),
      {variable: [operand, result]},
      [],
  )


def _infer_tiling_for_mma_ref(
    ref_ty: ir.MemRefType, max_swizzle: mgpu.SwizzlingMode
) -> tuple[int, int]:
  element_bytewidth = utils.bytewidth(ref_ty.element_type)
  strides, _ = ref_ty.get_strides_and_offset()
  min_dim_index = np.argmin(strides)
  minor_dim = ref_ty.shape[min_dim_index]

  # Try tiling with all swizzling modes starting from the largest one.
  for swizzle in [
      mgpu.SwizzlingMode.k128ByteSwizzle,
      mgpu.SwizzlingMode.k64ByteSwizzle,
      mgpu.SwizzlingMode.k32ByteSwizzle,
      mgpu.SwizzlingMode.kNoSwizzle,
  ]:
    if swizzle > max_swizzle:
      continue
    swizzle_elems = swizzle // element_bytewidth
    if minor_dim % swizzle_elems == 0:
      minor_tiling = swizzle_elems
      break
  else:
    # No valid tile transform can be inferred.
    raise ValueError(f"{ref_ty.shape} is not a valid WGMMA shape")

  major_tiling = 8
  transposed = min_dim_index != len(strides) - 1
  if transposed:
    tiling = (minor_tiling, major_tiling)
  else:
    tiling = (major_tiling, minor_tiling)
  return tiling


def _infer_wgmma_tiling(
    a_type: ir.Type, b_type: ir.MemRefType
) -> tuple[tuple[int, int] | None, tuple[int, int]]:
  """Infers the tiling for a (if in SMEM) and b of a WGMMAOp.

  If both a and b are in SMEM, this function infers tilings that have matching
  swizzle values.
  """
  b_tiling = _infer_tiling_for_mma_ref(
      b_type, max_swizzle=mgpu.SwizzlingMode.k128ByteSwizzle
  )
  b_swizzle = _compute_swizzle(b_type, lc.TileTransform(b_tiling))
  if not ir.MemRefType.isinstance(a_type):
    return None, b_tiling

  a_tiling = _infer_tiling_for_mma_ref(
      cast(ir.MemRefType, a_type), max_swizzle=b_swizzle
  )
  a_swizzle = _compute_swizzle(a_type, lc.TileTransform(a_tiling))
  if a_swizzle != b_swizzle:
    # The swizzle for a and b has to match. This is not a fundamental
    # limitation, rather the lowering doesn't currently support it.
    b_tiling = _infer_tiling_for_mma_ref(b_type, max_swizzle=a_swizzle)
    b_swizzle = _compute_swizzle(b_type, lc.TileTransform(b_tiling))
    assert a_swizzle == b_swizzle
  return a_tiling, b_tiling


@_add_equation_system_derivation_rule(mgpu.WGMMAOp)
def _wgmma_equation_system(
    ctx: DerivationContext,
    op: mgpu.WGMMAOp,
) -> tuple[eqns.EquationSystem, OperandOrResultsForVariable, list[Hint]]:
  assignments: dict[eqns.Variable, eqns.Constant] = {}
  # Registers
  vector_operands_or_results = vector_operands_and_results(op)
  vec_variable = eqns.Variable(vector_operands_or_results[0])
  assignments[vec_variable] = eqns.RegisterLayout(fa.WGMMA_LAYOUT)
  operands_or_results_for_variable = {vec_variable: vector_operands_or_results}

  # SMEM
  if ctx.enable_smem_inference:
    a_tiling, b_tiling = _infer_wgmma_tiling(op.a.type, op.b.type)
    b = OperandOrResult(op, VariableType.OPERAND, 2)
    b_var = ctx.producer_ref(b)

    assignments[b_var] = eqns.SMEMTiling(lc.TileTransform(b_tiling))
    operands_or_results_for_variable[b_var] = [b]

    if a_tiling is not None:
      # a is in SMEM
      a = OperandOrResult(op, VariableType.OPERAND, 1)
      a_var = ctx.producer_ref(a)
      assignments[a_var] = eqns.SMEMTiling(lc.TileTransform(a_tiling))
      operands_or_results_for_variable[a_var] = [a]

  system = eqns.EquationSystem(
      assignments=assignments,
  )
  return system, operands_or_results_for_variable, []


@_add_equation_system_derivation_rule(vector.BroadcastOp)
def _vector_broadcast_equation_system(
    ctx: DerivationContext,
    op: vector.BroadcastOp,
) -> tuple[eqns.EquationSystem, OperandOrResultsForVariable, list[Hint]]:
  del ctx
  # This is not expected to be necessary at the moment. We should be using
  # mgpu.BroadcastInDimOp instead when dealing with broadcasting vectors.
  if ir.ShapedType.isinstance(op.source.type):
    raise NotImplementedError("Only vector broadcasts from scalars are supported.")
  out_variable = eqns.Variable(OperandOrResult(op, VariableType.RESULT, 0))
  layout = eqns.RegisterLayout(fa.WGSplatFragLayout(tuple(op.result.type.shape)))
  return (
      eqns.EquationSystem(assignments={out_variable: layout}),
      {out_variable: [out_variable.key]},
      []
  )


@_add_equation_system_derivation_rule(vector.ReductionOp)
def _vector_reduction_equation_system(
    ctx: DerivationContext,
    op: vector.ReductionOp,
) -> tuple[eqns.EquationSystem, OperandOrResultsForVariable, list[Hint]]:
  del ctx
  in_variable = eqns.Variable(OperandOrResult(op, VariableType.OPERAND, 0))
  return eqns.EquationSystem(), {in_variable: [in_variable.key]}, []


def _reduction_equation_and_hint(
    larger: eqns.Variable,
    smaller: eqns.Variable,
    larger_shape: tuple[int, ...],
    reduction_dims: tuple[int, ...]
) -> tuple[eqns.Equation, Hint]:
  reduce_expr = eqns.Reduce(larger, reduction_dims)
  # There are always many options for broadcasting a layout, so we can only
  # derive a broadcast hint in the out_variable -> source_variable direction.
  broadcast_dims = tuple(
      i for i in range(len(larger_shape)) if i not in reduction_dims
  )
  broadcast_expr = eqns.BroadcastInDim(smaller, broadcast_dims, larger_shape)
  broadcast_hint = Hint(variable=larger, expression=broadcast_expr)
  return eqns.Equation(lhs=smaller, rhs=reduce_expr), broadcast_hint


@_add_equation_system_derivation_rule(vector.MultiDimReductionOp)
def _multi_dim_reduction_equation_system(
    ctx: DerivationContext,
    op: vector.MultiDimReductionOp,
) -> tuple[eqns.EquationSystem, OperandOrResultsForVariable, list[Hint]]:
  del ctx
  source = OperandOrResult(op, VariableType.OPERAND, 0)
  acc = OperandOrResult(op, VariableType.OPERAND, 1)
  out = OperandOrResult(op, VariableType.RESULT, 0)
  source_variable = eqns.Variable(source)
  out_variable = eqns.Variable(out)

  reduction_equation, broadcast_hint = _reduction_equation_and_hint(
      source_variable, out_variable,
      tuple(ir.ShapedType(op.source.type).shape), tuple(op.reduction_dims)
  )
  # TODO(bchetioui): in the future, we may need to add rules that prevent
  # strided layouts from being chosen---since trying to reduce a strided layout
  # may cause us to raise an Exception at the moment.
  return (
      eqns.EquationSystem(equations=[reduction_equation]),
      {source_variable: [source], out_variable: [acc, out]},
      [broadcast_hint],
  )


@_add_equation_system_derivation_rule(mgpu.BroadcastInDimOp)
def _broadcast_in_dim_equation_system(
    ctx: DerivationContext,
    op: mgpu.BroadcastInDimOp,
) -> tuple[eqns.EquationSystem, OperandOrResultsForVariable, list[Hint]]:
  del ctx
  out_variable = eqns.Variable(OperandOrResult(op, VariableType.RESULT, 0))
  source_variable = eqns.Variable(OperandOrResult(op, VariableType.OPERAND, 0))
  out_shape = tuple(cast(ir.ShapedType, op.result.type).shape)
  reduction_dims = tuple(
      i for i in range(len(out_shape)) if i not in op.broadcast_dimensions
  )

  reduction_equation, broadcast_hint = _reduction_equation_and_hint(
      out_variable, source_variable, out_shape, reduction_dims
  )

  return (
      eqns.EquationSystem(equations=[reduction_equation]),
      {source_variable: [source_variable.key], out_variable: [out_variable.key]},
      [broadcast_hint],
  )


@_add_equation_system_derivation_rule(vector.ShapeCastOp)
def _shape_cast_equation_system(
    ctx: DerivationContext, op: vector.ShapeCastOp
) -> tuple[eqns.EquationSystem, OperandOrResultsForVariable, list[Hint]]:
  del ctx
  in_shape = tuple(cast(ir.ShapedType, op.source.type).shape)
  out_shape = tuple(cast(ir.ShapedType, op.result.type).shape)

  in_variable = eqns.Variable(OperandOrResult(op, VariableType.OPERAND, 0))
  out_variable = eqns.Variable(OperandOrResult(op, VariableType.RESULT, 0))

  # Here, we are in a case where we are stating
  #
  #   out_variable = reshape(in_variable, in_shape, out_shape).
  #
  # Thanks to the symmetric property of reshape, we can also issue an equation
  # in the other direction, i.e.
  #
  #   in_variable = reshape(out_variable, out_shape, in_shape)
  #
  # in order to be able to figure out an assignment for `in_variable`. if we
  # happen to know `out_variable`. If we only issue the first equation, then
  # we will not be able to figure out an assignment for `in_variable` if we
  # only know `out_variable`, even though their relationship is fully
  # determined.
  in_to_out = eqns.Reshape(
      in_variable, source_shape=in_shape, target_shape=out_shape
  )
  out_to_in = eqns.Reshape(
      out_variable, source_shape=out_shape, target_shape=in_shape
  )

  return (
      eqns.EquationSystem(
          equations=[
              eqns.Equation(lhs=out_variable, rhs=in_to_out),
              eqns.Equation(lhs=in_variable, rhs=out_to_in),
          ],
      ),
      {in_variable: [in_variable.key], out_variable: [out_variable.key]},
      [],
  )


@_add_equation_system_derivation_rule(mgpu.CustomPrimitiveOp)
def _custom_primitive_equation_system(
    ctx: DerivationContext,
    op: mgpu.CustomPrimitiveOp,
) -> tuple[eqns.EquationSystem, OperandOrResultsForVariable, list[Hint]]:
  assignments: dict[eqns.Variable, eqns.Constant] = {}
  equations: list[eqns.Equation] = []
  in_layouts = iter(op.in_layouts)
  in_transforms = iter(op.in_transforms)
  variables: list[eqns.Variable] = []
  for i, operand in enumerate(op.operands):
    if is_vector(operand):
      v = eqns.Variable(OperandOrResult(op, VariableType.OPERAND, i))
      variables.append(v)
      assignments[v] = eqns.RegisterLayout(
          layouts_lib.from_layout_attr(next(in_layouts))
      )
    elif ctx.enable_smem_inference and _is_smem_ref(operand):
      # Here we need to create a new variable, even though it is equal to the
      # source operand. This is because we directly assign the new variable and
      # if we did that to the source there could be conflicting assignments.
      # For example, the same ref could be passed into the custom op twice with
      # different transforms, which needs to yield an unsatisfiable system.
      #
      # TODO(b/447079781): Consider creating the final Equation system using
      # __and__ and potentially returning Unsatisfiable() directly if there is
      # a conflict between the assignments.
      operand_or_result = OperandOrResult(op, VariableType.OPERAND, i)
      source_var = ctx.producer_ref(operand_or_result)
      v = eqns.Variable(operand_or_result)
      equations.append(eqns.Equation(lhs=source_var, rhs=v))
      variables.append(v)
      transforms = next(in_transforms)
      ref_ty = operand_or_result.value.type
      tiling = _extract_smem_tiling_from_custom_transform_attrs(ref_ty, transforms)
      assignments[v] = tiling

  out_layouts = iter(op.out_layouts)
  for i, result in enumerate(op.results):
    if ir.VectorType.isinstance(result.type):
      v = eqns.Variable(OperandOrResult(op, VariableType.RESULT, i))
      variables.append(v)
      assignments[v] = eqns.RegisterLayout(
          layouts_lib.from_layout_attr(next(out_layouts))
      )
  return (
      eqns.EquationSystem(equations=equations, assignments=assignments),
      {v: [v.key] for v in variables},
      [],
  )


def _tmem_layout_from_layout_attr(
    layout_attr: mgpu.TiledLayout,
) -> tcgen05.TMEMLayout:
  layout = layouts_lib.from_layout_attr(layout_attr)
  assert isinstance(layout, fa.TiledLayout)
  return tcgen05.TMEMLayout(
      layout.tiling, layout.warp_dims, layout.lane_dims, layout.vector_dim
  )


@_add_equation_system_derivation_rule(mgpu.TmemLayoutCastOp)
def _tmem_layout_cast_equation_system(
    ctx: DerivationContext,
    op: mgpu.TmemLayoutCastOp,
) -> tuple[eqns.EquationSystem, OperandOrResultsForVariable, list[Hint]]:
  operand = OperandOrResult(op, VariableType.OPERAND, 0)
  variable = ctx.producer_ref(operand)
  result = OperandOrResult(op, VariableType.RESULT, 0)
  out_layout = eqns.TMEMLayout(_tmem_layout_from_layout_attr(op.new_layout))
  return (
      eqns.EquationSystem(assignments={variable: out_layout}),
      {variable: [operand, result]},
      [],
  )


@_add_equation_system_derivation_rule(mgpu.TmemAllocOp)
def _tmem_alloc_equation_system(
    ctx: DerivationContext,
    op: mgpu.TmemAllocOp,
) -> tuple[eqns.EquationSystem, OperandOrResultsForVariable, list[Hint]]:
  del ctx
  result = OperandOrResult(op, VariableType.RESULT, 0)
  result_var = eqns.Variable(result)
  layout = tcgen05._infer_tmem_layout(
      tuple(op.result.type.shape), op.collective, packing=1
  )

  in_smem = OperandOrResult(op, VariableType.OPERAND, 0)
  in_smem_var = eqns.Variable(in_smem)
  assignments: dict[eqns.Variable, eqns.Constant] = {
      in_smem_var: eqns.SMEMTiling(None)
  }
  operands_for_variable = {result_var: [result], in_smem_var: [in_smem]}

  # This is a hint, not a hard constraint. This will be the default layout if
  # none can be inferred.
  hint = Hint(result_var, eqns.TMEMLayout(layout))
  system = eqns.EquationSystem(assignments=assignments)
  return system, operands_for_variable, [hint]


@_add_equation_system_derivation_rule(mgpu.TmemDeallocOp)
def _tmem_dealloc_equation_system(
    ctx: DerivationContext,
    op: mgpu.TmemDeallocOp,
) -> tuple[eqns.EquationSystem, OperandOrResultsForVariable, list[Hint]]:
  operand = OperandOrResult(op, VariableType.OPERAND, 0)
  variable = ctx.producer_ref(operand)
  return eqns.EquationSystem(), {variable: [operand]}, []


@_add_equation_system_derivation_rule(mgpu.TcGen05MMAOp)
def _tcgen05_mma_equation_system(
    ctx: DerivationContext,
    op: mgpu.TcGen05MMAOp,
) -> tuple[eqns.EquationSystem, OperandOrResultsForVariable, list[Hint]]:
  assignments: dict[eqns.Variable, eqns.Constant] = {}
  operands_for_variable: OperandOrResultsForVariable = {}

  # TMEM
  acc = OperandOrResult(op, VariableType.OPERAND, 0)
  acc_variable = ctx.producer_ref(acc)
  acc_type = ir.ShapedType(op.accumulator.type)
  acc_layout = tcgen05._infer_tmem_layout(
      tuple(acc_type.shape), op.collective, packing=1
  )
  assignments[acc_variable] = eqns.TMEMLayout(acc_layout)
  operands_for_variable[acc_variable] = [acc]

  if _is_tmem_ref(op.a):
    a = OperandOrResult(op, VariableType.OPERAND, 1)
    a_type = ir.ShapedType(op.a.type)
    a_var = ctx.producer_ref(a)
    packing = 32 // utils.bitwidth(a_type.element_type)
    a_layout = tcgen05._infer_tmem_layout(
        tuple(a_type.shape), op.collective, packing
    )
    assignments[a_var] = eqns.TMEMLayout(a_layout)
    operands_for_variable[a_var] = [a]

  # SMEM
  if ctx.enable_smem_inference:
    b_tiling = _infer_tiling_for_mma_ref(
        ir.MemRefType(op.b.type),
        max_swizzle=mgpu.SwizzlingMode.k128ByteSwizzle,
    )
    b = OperandOrResult(op, VariableType.OPERAND, 2)
    b_var = ctx.producer_ref(b)
    assignments[b_var] = eqns.SMEMTiling(lc.TileTransform(b_tiling))
    operands_for_variable[b_var] = [b]

    if _is_smem_ref(op.a):
      a_tiling = _infer_tiling_for_mma_ref(
          ir.MemRefType(op.a.type),
          max_swizzle=mgpu.SwizzlingMode.k128ByteSwizzle,
      )
      a = OperandOrResult(op, VariableType.OPERAND, 1)
      a_var = ctx.producer_ref(a)
      assignments[a_var] = eqns.SMEMTiling(lc.TileTransform(a_tiling))
      operands_for_variable[a_var] = [a]

  return eqns.EquationSystem(assignments=assignments), operands_for_variable, []


@_add_equation_system_derivation_rule(mgpu.AsyncLoadTmemOp)
def _async_load_tmem_equation_system(
    ctx: DerivationContext,
    op: mgpu.AsyncLoadTmemOp,
) -> tuple[eqns.EquationSystem, OperandOrResultsForVariable, list[Hint]]:
  source = OperandOrResult(op, VariableType.OPERAND, 0)
  source_variable = ctx.producer_ref(source)
  destination = OperandOrResult(op, VariableType.RESULT, 0)
  destination_variable = eqns.Variable(destination)
  constraint = eqns.IsTransferable(
      source_variable,
      destination_variable,
      tuple(ir.ShapedType(op.source.type).shape),
  )
  return (
      eqns.EquationSystem(constraints=[constraint]),
      {source_variable: [source], destination_variable: [destination]},
      [],
  )


@_add_equation_system_derivation_rule(mgpu.AsyncStoreTmemOp)
def _async_store_tmem_equation_system(
    ctx: DerivationContext,
    op: mgpu.AsyncStoreTmemOp,
) -> tuple[eqns.EquationSystem, OperandOrResultsForVariable, list[Hint]]:
  source = OperandOrResult(op, VariableType.OPERAND, 0)
  source_variable = eqns.Variable(source)
  destination = OperandOrResult(op, VariableType.OPERAND, 1)
  destination_variable = ctx.producer_ref(destination)
  constraint = eqns.IsTransferable(
      source_variable,
      destination_variable,
      tuple(ir.ShapedType(op.source.type).shape),
  )
  return (
      eqns.EquationSystem(constraints=[constraint]),
      {source_variable: [source], destination_variable: [destination]},
      [],
  )


@_add_equation_system_derivation_rule(mgpu.SliceSMEMOp)
def _slice_smem_equation_system(
    ctx: DerivationContext,
    op: mgpu.SliceSMEMOp,
) -> tuple[eqns.EquationSystem, OperandOrResultsForVariable, list[Hint]]:
  del ctx
  res = OperandOrResult(op, VariableType.RESULT, 0)
  res_var = eqns.Variable(res)
  return (eqns.EquationSystem(), {res_var: [res]}, [])


# TODO(b/447079781): Check whether we still need this rule. Normally,
# DynamicSharedMemory is only generated in the lowering pass. If there is
# another case where a ViewOp is generated beforehand, and this rule cannot be
# removed, document here what that case is.
@_add_equation_system_derivation_rule(memref.ViewOp)
def _memref_view_op_equation_system(
    ctx: DerivationContext,
    op: memref.ViewOp,
) -> tuple[eqns.EquationSystem, OperandOrResultsForVariable, list[Hint]]:
  del ctx

  # The source is expected to come from a DynamicSharedMemoryOp which does not
  # participate in layout inference and no variable exists for it.
  if not isinstance(op.source.owner.opview, gpu.DynamicSharedMemoryOp):
    raise NotImplementedError(
        "Memref view transforms are only inferred when the op is a direct user "
        f"of a DynamicSharedMemoryOp but got {op}."
    )

  res = OperandOrResult(op, VariableType.RESULT, 0)
  res_var = eqns.Variable(res)
  return eqns.EquationSystem(), {res_var: [res]}, []


# TODO(b/447079781): Check whether we should create new variables or use
# variables from the producer_ref in all memref rules. Ideally all of those
# rules should handles references the same way. If there is a reason for
# handling them differently, document in each rule.
#
# E.g. here, using the producer_ref variable would work for all existing tests,
# but we create a new variable + an equation in order to be consistent with the
# other memref rules.
@_add_equation_system_derivation_rule(memref.CastOp)
def _memref_cast_op_equation_system(
    ctx: DerivationContext,
    op: memref.CastOp,
) -> tuple[eqns.EquationSystem, OperandOrResultsForVariable, list[Hint]]:
  source = OperandOrResult(op, VariableType.OPERAND, 0)
  var_source = ctx.producer_ref(source)
  dest = OperandOrResult(op, VariableType.RESULT, 0)
  var_dest = eqns.Variable(dest)
  return (
      eqns.EquationSystem(equations=[eqns.Equation(var_source, var_dest)]),
      {var_source: [source], var_dest: [dest]},
      [],
  )


@_add_equation_system_derivation_rule(memref.TransposeOp)
def _memref_transpose_op_equation_system(
    ctx: DerivationContext,
    op: memref.TransposeOp,
) -> tuple[eqns.EquationSystem, OperandOrResultsForVariable, list[Hint]]:
  in_ty = ir.MemRefType(op.in_.type)
  if len(in_ty.shape) != 2:
    raise NotImplementedError(f"Only 2D memrefs are supported, got {in_ty}")
  in_strides, _ = in_ty.get_strides_and_offset()
  out_strides, _ = ir.MemRefType(op.result.type).get_strides_and_offset()
  transpose = in_strides != out_strides

  source = OperandOrResult(op, VariableType.OPERAND, 0)
  dest = OperandOrResult(op, VariableType.RESULT, 0)
  source_var = ctx.producer_ref(source)

  if not transpose:
    return (eqns.EquationSystem(), {source_var: [source, dest]}, [])

  dest_var = eqns.Variable(dest)
  equations = [
      eqns.Equation(source_var, eqns.Transpose(dest_var)),
      eqns.Equation(eqns.Transpose(source_var), dest_var),
  ]
  system = eqns.EquationSystem(equations=equations)
  return system, {source_var: [source], dest_var: [dest]}, []


# `memref.load` and `memref.store` are used to load barrier phases which are
# scalars---the rule needn't do anything interesting, but we need to have it.
@_add_equation_system_derivation_rule(memref.LoadOp)
@_add_equation_system_derivation_rule(memref.StoreOp)
def _memref_load_store_op_equation_system(
    ctx: DerivationContext,
    op: memref.LoadOp | memref.StoreOp,
) -> tuple[eqns.EquationSystem, OperandOrResultsForVariable, list[Hint]]:
  del ctx

  ref_shape = ir.MemRefType(op.memref.type).shape
  if ref_shape and ref_shape != [1]:
    raise NotImplementedError(
        f"Only scalar memrefs are supported, got {ref_shape}"
    )

  ref_op_index = 0 if isinstance(op, memref.LoadOp) else 1
  ref = OperandOrResult(op, VariableType.OPERAND, ref_op_index)
  var = eqns.Variable(ref)
  assignments: dict[eqns.Variable, eqns.Constant] = {var: eqns.SMEMTiling(None)}
  return eqns.EquationSystem(assignments=assignments), {var: [ref]}, []


def _extract_smem_tiling_from_custom_transform_attrs(
    ref_type: ir.MemRefType,
    transform_attrs: ir.ArrayAttr,
) -> eqns.SMEMTiling:
  transforms = [layouts_lib.from_transform_attr(x) for x in transform_attrs]
  match transforms:
    case []:
      tile_transform = None
      swizzle = None
    case [lc.TileTransform() as t]:
      tile_transform = t
      swizzle = None
    case [lc.TileTransform() as t, mgpu.SwizzlingMode() as s]:
      tile_transform = t
      swizzle = s
    case _:
      raise NotImplementedError(f"Unsupported transforms {transforms}")

  if swizzle is not None:
    computed_swizzle = _compute_swizzle(ref_type, tile_transform)
    if computed_swizzle != swizzle:
      raise NotImplementedError(
          f"Cannot honor caller-provided swizzle {swizzle} that is different "
          f"from the computed swizle {computed_swizzle} for type {ref_type}."
      )

  return eqns.SMEMTiling(tile_transform)


@_add_equation_system_derivation_rule(mgpu.WithTransformsOp)
def _with_transforms_equation_system(
    ctx: DerivationContext,
    op: mgpu.WithTransformsOp,
) -> tuple[eqns.EquationSystem, OperandOrResultsForVariable, list[Hint]]:
  source = OperandOrResult(op, VariableType.OPERAND, 0)
  dest = OperandOrResult(op, VariableType.RESULT, 0)
  var = ctx.producer_ref(source)
  tiling = _extract_smem_tiling_from_custom_transform_attrs(op.ref.type, op.transforms)
  assignments: dict[eqns.Variable, eqns.Constant] = {var: tiling}
  return eqns.EquationSystem(assignments=assignments), {var: [source, dest]}, []


@_add_equation_system_derivation_rule(mgpu.AsyncLoadOp)
@_add_equation_system_derivation_rule(mgpu.AsyncStoreOp)
def _async_load_store_equation_system(
    ctx: DerivationContext,
    op: mgpu.AsyncLoadOp | mgpu.AsyncStoreOp,
) -> tuple[eqns.EquationSystem, OperandOrResultsForVariable, list[Hint]]:
  operand_index = 1 if isinstance(op, mgpu.AsyncLoadOp) else 0
  operand = OperandOrResult(op, VariableType.OPERAND, operand_index)
  var = ctx.producer_ref(operand)
  return eqns.EquationSystem(), {var: [operand]}, []


def _ensure_all_layouts_are_set(
    op: ir.OpView, enable_smem_inference: bool
) -> None:
  if inference_utils.should_have_layout(op):
    _ensure_right_number_of_layouts(
        op,
        inference_utils.in_layouts(op)
        if inference_utils.has_in_layouts_set(op)
        else [],
        inference_utils.out_layouts(op)
        if inference_utils.has_out_layouts_set(op)
        else [],
    )
  if inference_utils.should_have_tmem_layout(op):
    _ensure_right_number_of_tmem_layouts(
        op,
        inference_utils.in_tmem_layouts(op)
        if inference_utils.has_in_tmem_layouts_set(op)
        else [],
        inference_utils.out_tmem_layouts(op)
        if inference_utils.has_out_tmem_layouts_set(op)
        else [],
    )
  if enable_smem_inference and inference_utils.should_have_transforms(op):
    _ensure_right_number_of_transforms(
        op,
        inference_utils.in_transforms(op)
        if inference_utils.has_in_transforms_set(op)
        else [],
        inference_utils.out_transforms(op)
        if inference_utils.has_out_transforms_set(op)
        else [],
    )


def _ensure_right_number_of_layouts(
    op: ir.OpView,
    in_layouts: Sequence[fa.FragmentedLayout | ir.Attribute],
    out_layouts: Sequence[fa.FragmentedLayout | ir.Attribute],
) -> None:
  """Ensures that the right number of in/out layouts are provided for an op."""
  if len(in_layouts) != sum(map(is_vector, op.operands)):
    raise ValueError(
        "Expected the same number of in_layouts as vector operands."
    )
  if len(out_layouts) != sum(map(is_vector, op.results)):
    raise ValueError(
        "Expected the same number of out_layouts as vector results."
    )


def _ensure_right_number_of_tmem_layouts(
    op: ir.OpView,
    in_layouts: Sequence[ir.Attribute],
    out_layouts: Sequence[ir.Attribute],
) -> None:
  """Ensures that the right number of in/out TMEM layouts are provided for an op."""
  if len(in_layouts) != sum(map(_is_tmem_ref, op.operands)):
    raise ValueError(
        f"Expected the same number of in_tmem_layouts({in_layouts}) as TMEM ref operands. op=\n  {op}"
    )
  if len(out_layouts) != sum(map(_is_tmem_ref, op.results)):
    raise ValueError(
        f"Expected the same number of out_tmem_layouts({out_layouts}) as TMEM ref results. op=\n  {op}"
    )


def _ensure_right_number_of_transforms(
    op: ir.OpView,
    in_transforms: Sequence[Any],
    out_transforms: Sequence[Any],
) -> None:
  """Ensures that the right number of in/out SMEM transforms are provided for an op."""
  if len(in_transforms) != sum(
      map(inference_utils.is_transformable_smem_memref, op.operands)
  ):
    raise ValueError(
        f"Expected the same number of in_transforms({in_transforms}) as SMEM ref operands. op=\n  {op}"
    )
  if len(out_transforms) != sum(
      map(inference_utils.is_transformable_smem_memref, op.results)
  ):
    raise ValueError(
        f"Expected the same number of out_transforms({out_transforms}) as SMEM ref results. op=\n  {op}"
    )


def _compute_swizzle(
    type: ir.Type, tile_transform: lc.TileTransform | None
) -> mgpu.SwizzlingMode:
  """Computes the swizzle mode given a tiling transform and a data type."""
  if tile_transform is None:
    # TODO(b/447079781): Revisit if this is the behavior we want.
    return mgpu.SwizzlingMode.kNoSwizzle

  if not ir.MemRefType.isinstance(type):
    raise ValueError(f"Expected a MemRefType, got {type}.")
  ref_ty = ir.MemRefType(type)
  strides, _ = ref_ty.get_strides_and_offset()
  tiling = tile_transform.tiling

  if len(tiling) > len(strides):
    raise ValueError(
        f"The tile rank ({len(tiling)}) cannot be greater than the ref's rank"
        f" ({len(strides)})."
    )

  minor_tiling = tiling[np.argmin(strides[-len(tiling):])]
  swizzle = minor_tiling * utils.bytewidth(ref_ty.element_type)
  assert swizzle in (
      mgpu.SwizzlingMode.k128ByteSwizzle,
      mgpu.SwizzlingMode.k64ByteSwizzle,
      mgpu.SwizzlingMode.k32ByteSwizzle,
      mgpu.SwizzlingMode.kNoSwizzle,
  )
  return mgpu.SwizzlingMode(swizzle)


@dataclasses.dataclass(frozen=True)
class _TypeAndLayout:
  type: ir.Type
  layout: eqns.Constant


def assign_layouts(
    solution: dict[OperandOrResult, eqns.Constant],
    enable_smem_inference: bool,
) -> None:
  """Assigns the layouts in `solution` to the MLIR ops they belong to.

  This function requires that, for each MLIR op that appears in `solution`,
  `solution` contains a layout assignment for all of its `vector`, TMEM, and
  SMEM operands and results.
  """
  solution_sorted_by_op = sorted(
      solution.items(), key=lambda kv: id(kv[0].operation)
  )
  solution_per_op = itertools.groupby(
      solution_sorted_by_op, key=lambda kv: kv[0].operation
  )

  for op, assignments in solution_per_op:
    assignments_sorted_by_type = sorted(assignments, key=lambda kv: kv[0].type)
    assignments_by_type = {
        ty: list(group)
        for ty, group in itertools.groupby(
            assignments_sorted_by_type, key=lambda kv: kv[0].type
        )
    }

    in_assignments = assignments_by_type.get(VariableType.OPERAND, [])
    out_assignments = assignments_by_type.get(VariableType.RESULT, [])

    index = lambda kv: kv[0].index
    in_tls = [
        _TypeAndLayout(v.value.type, ce)
        for v, ce in sorted(in_assignments, key=index)
    ]
    out_tls = [
        _TypeAndLayout(v.value.type, ce)
        for v, ce in sorted(out_assignments, key=index)
    ]

    in_layouts = [
        tl.layout.value for tl in in_tls if isinstance(tl.layout, eqns.RegisterLayout)
    ]
    out_layouts = [
        tl.layout.value for tl in out_tls if isinstance(tl.layout, eqns.RegisterLayout)
    ]
    in_tmem_layouts = [
        tl.layout.value for tl in in_tls if isinstance(tl.layout, eqns.TMEMLayout)
    ]
    out_tmem_layouts = [
        tl.layout.value for tl in out_tls if isinstance(tl.layout, eqns.TMEMLayout)
    ]
    in_transforms = [tl for tl in in_tls if isinstance(tl.layout, eqns.SMEMTiling)]
    out_transforms = [tl for tl in out_tls if isinstance(tl.layout, eqns.SMEMTiling)]

    _ensure_right_number_of_layouts(op, in_layouts, out_layouts)
    _ensure_right_number_of_tmem_layouts(op, in_tmem_layouts, out_tmem_layouts)
    if enable_smem_inference:
      _ensure_right_number_of_transforms(op, in_transforms, out_transforms)

    if inference_utils.should_have_in_layout(op):
      attrs = [layouts_lib.to_layout_attr(l) for l in in_layouts]
      op.attributes["in_layouts"] = ir.ArrayAttr.get(attrs)
    if inference_utils.should_have_out_layout(op):
      attrs = [layouts_lib.to_layout_attr(l) for l in out_layouts]
      op.attributes["out_layouts"] = ir.ArrayAttr.get(attrs)
    if inference_utils.should_have_in_tmem_layout(op):
      attrs = [layouts_lib.to_layout_attr(l) for l in in_tmem_layouts]
      op.attributes["in_tmem_layouts"] = ir.ArrayAttr.get(attrs)
    if inference_utils.should_have_out_tmem_layout(op):
      attrs = [layouts_lib.to_layout_attr(l) for l in out_tmem_layouts]
      op.attributes["out_tmem_layouts"] = ir.ArrayAttr.get(attrs)

    if enable_smem_inference:

      def _to_transform_attrs(
          transforms: list[_TypeAndLayout],
      ) -> list[ir.ArrayAttr]:
        all_attrs: list[ir.ArrayAttr] = []
        for tl in transforms:
          assert isinstance(tl.layout, eqns.SMEMTiling)  # make pytype happy
          attrs = []
          if tl.layout.value is not None:
            attrs.append(layouts_lib.to_transform_attr(tl.layout.value))
            swizzle = _compute_swizzle(tl.type, tl.layout.value)
            attrs.append(layouts_lib.to_transform_attr(swizzle))
          all_attrs.append(ir.ArrayAttr.get(attrs))
        return all_attrs

      if inference_utils.should_have_in_transforms(op):
        attrs = _to_transform_attrs(in_transforms)
        op.attributes["in_transforms"] = ir.ArrayAttr.get(attrs)
      if inference_utils.should_have_out_transforms(op):
        attrs = _to_transform_attrs(out_transforms)
        op.attributes["out_transforms"] = ir.ArrayAttr.get(attrs)


def vector_operands_and_results(op: ir.OpView) -> list[OperandOrResult]:
  """Returns all the vector operands and results for the given op."""
  operands_or_results = [
      OperandOrResult(op, VariableType.OPERAND, i)
      for i, o in enumerate(op.operands)
      if is_vector(o)
  ]
  operands_or_results.extend([
      OperandOrResult(op, VariableType.RESULT, i)
      for i, o in enumerate(op.results)
      if is_vector(o)
  ])
  return operands_or_results


def producer_result(operand: OperandOrResult) -> OperandOrResult:
  """Given an operand, returns the corresponding result in its producer.

  When the producer is a block, we return the corresponding operand in the
  operation that owns the block.
  """
  assert operand.type == VariableType.OPERAND
  value = operand.operation.operands[operand.index]
  producer = value.owner
  if isinstance(producer, ir.Operation):
    index = list(producer.results).index(value)
    return OperandOrResult(producer.opview, VariableType.RESULT, index)

  # Block case, useful for deriving layouts for ops
  # depending on function parameters, or loop block arguments.
  if isinstance(producer, ir.Block):
    index = list(cast(ir.Block, producer).arguments).index(value)
    if isinstance(producer.owner, scf.ForOp):
      # In this case, the block arguments are offset compared to the loop
      # operands. The loop operands have the lower bound, upper bound, and step
      # as their leading arguments. The block arguments omit these parameters,
      # but start with the iteration variable.
      num_leading_args = 3
      index += num_leading_args - 1
      return OperandOrResult(producer.owner.opview, VariableType.OPERAND, index)
    if isinstance(producer.owner, scf.WhileOp):
      [before_block] = producer.owner.before.blocks
      [after_block] = producer.owner.after.blocks
      if producer == before_block:
        # In this case, the block arguments correspond to the while operands.
        return OperandOrResult(producer.owner.opview, VariableType.OPERAND, index)
      else:
        assert producer == after_block
        # In this case, the block arguments correspond to the while results.
        return OperandOrResult(producer.owner.opview, VariableType.RESULT, index)
    raise NotImplementedError(
        f"Producer {producer} is not a ForOp, a WhileOp: {type(producer)}."
    )

  raise TypeError(
      f"Producer {producer} is not an operation nor a block: {type(producer)}."
  )


def consumer_operands(result: OperandOrResult) -> Sequence[OperandOrResult]:
  """Given a result, returns the corresponding operands in its consumers."""
  assert result.type == VariableType.RESULT
  consumer_operands: list[OperandOrResult] = []
  # The layout can also be chosen from the layout of the consumers of the
  # results.
  for use in cast(ir.OpResult, result.operation.results[result.index]).uses:
    consumer = use.owner.opview  # pytype: disable=attribute-error
    index = use.operand_number
    consumer_operands.append(OperandOrResult(consumer, VariableType.OPERAND, index))
  return consumer_operands


def derive_hints_and_constraints(
    operands_and_results_for_variable: OperandOrResultsForVariable
) -> tuple[list[Hint], list[eqns.Relayout]]:
  """Derives propagation hints from the given variable mapping."""
  hints: list[Hint] = []
  constraints: list[eqns.Relayout] = []
  variable_for_operand_or_result: dict[OperandOrResult, eqns.Variable] = {}
  for variable, operand_and_results in operands_and_results_for_variable.items():
    for operand_or_result in operand_and_results:
      if operand_or_result in variable_for_operand_or_result:
        raise ValueError(
            f"{operand_or_result} is mapped to both {variable} and "
            f"{variable_for_operand_or_result[operand_or_result]}"
        )
    variable_for_operand_or_result |= {k: variable for k in operand_and_results}

  visited: set[eqns.Variable] = set()
  for variable, operand_and_results in operands_and_results_for_variable.items():
    producers: list[eqns.Variable] = []
    consumers: list[eqns.Variable] = []
    for operand_or_result in operand_and_results:
      # We can only relayout variables that are in registers.
      if operand_or_result.memory_space != MemorySpace.REG:
        continue

      if operand_or_result.type == VariableType.OPERAND:
        pr = producer_result(operand_or_result)
        producer_variable = variable_for_operand_or_result[pr]
        producers.append(producer_variable)
        # Only add the constraint if we haven't already created that constraint
        # when processing this variable as one of the producer's consumers.
        if producer_variable not in visited:
          # The producer of a variable must be relayout-able to the variable.
          constraints.append(eqns.Relayout(producer_variable, variable))
      elif operand_or_result.type == VariableType.RESULT:
        for co in consumer_operands(operand_or_result):
          consumer_variable = variable_for_operand_or_result[co]
          consumers.append(consumer_variable)
          # Only add the constraint if we haven't already created that
          # constraint when processing this variable as the consumer's producer.
          if consumer_variable not in visited:
            # A variable must be relayout-able to its consumers.
            constraints.append(eqns.Relayout(variable, consumer_variable))
    visited.add(variable)

    if producers:
      least_replicated_producer = eqns.LeastReplicated(tuple(producers))
      hint_expr = eqns.MostReplicated((least_replicated_producer, *consumers))
      hints.append(Hint(variable, hint_expr))
    elif consumers:
      hint_expr = eqns.MostReplicated(tuple(consumers))
      hints.append(Hint(variable, hint_expr))

  return hints, constraints


def is_terminator(op: ir.OpView) -> bool:
  return isinstance(op, (scf.YieldOp, scf.ConditionOp))


def _drop_smem(
    system: eqns.EquationSystem, ctx: DerivationContext
) -> tuple[eqns.EquationSystem, DerivationContext]:
  """Drops SMEM related variables constraints and hints.

  This is only needed to enable the gradual implementation and testing of
  SMEM inference.
  TODO(b/447079781): Remove this function once SMEM inference is fully
  implemented.
  """
  def is_smem(
      x: (
          OperandOrResult
          | eqns.Constraint
          | eqns.Expression
          | eqns.Equation
      ),
  ) -> bool:
    match x:
      case OperandOrResult(memory_space=MemorySpace.SMEM):
        return True
      case eqns.Transpose():
        return True
      case eqns.IsTransferable(source=source, target=target):
        return is_smem(source) or is_smem(target)
      case eqns.Divides():
        return True
      case eqns.Variable(key=key):
        return is_smem(key)
      case eqns.SMEMTiling():
        return True
      case eqns.Equation(lhs=lhs, rhs=rhs):
        return is_smem(lhs) or is_smem(rhs)
      case _:
        return False

  assign = {k: v for k, v in system.assignments.items() if not is_smem(v)}
  equations = [e for e in system.equations if not is_smem(e)]
  const = [c for c in system.constraints if not is_smem(c)]

  new_ctx = DerivationContext()

  new_ctx.operand_and_results_for_variable = {
      k: v for k, v in ctx.operand_and_results_for_variable.items() if not is_smem(k)
  }

  new_ctx.variable_for_operand_or_result = {
      k: v for k, v in ctx.variable_for_operand_or_result.items() if not is_smem(k)
  }

  return eqns.EquationSystem(assign, equations, const), new_ctx


def infer_layout(module: ir.Module, enable_smem_inference: bool = False):
  """Infers layouts for the given module.

  * If there are vector (respectively SMEM refs, TMEM refs) operands,
  `in_layouts` (respectively `in_transforms`, `in_tmem_layouts`) will be set and
  contain one element per relevant argument in the memory space.
  * If there are vector (respectively SMEM refs, TMEM refs) outputs,
  `out_layouts` (respectively `out_transforms`, `out_tmem_layouts`) will be set
  and contain one element per relevant argument in the memory space.
  * Any of these attributes is guaranteed to not be set if there is no relevant
  input/output in the corresponding memory space.

  If `enable_smem_inference` is False, SMEM transforms are not inferred. This
  is only a temporary flag to allow for an incremental rollout of SMEM
  inference.
  TODO(b/447079781): Remove this flag once SMEM inference is fully implemented.
  """
  global_equation_system: eqns.EquationSystem | eqns.Unsatisfiable
  global_equation_system = eqns.EquationSystem()
  hints: list[Hint] = []
  ctx = DerivationContext(enable_smem_inference=enable_smem_inference)

  def gather_equations(op: ir.Operation):
    # Terminator ops are handled directly by the op whose region they belong to.
    # This is because they need to be in sync with their parent op's inputs and
    # outputs---and the parent op's equations therefore need to take them them
    # into account.
    if is_terminator(op):
      return
    should_have_layout = (
        inference_utils.should_have_layout(op)
        or inference_utils.should_have_tmem_layout(op)
        or (inference_utils.should_have_transforms(op) and enable_smem_inference)
    )
    if not should_have_layout:
      return
    rule = _equation_system_derivation_rules.get(op.OPERATION_NAME, None)  # pytype: disable=attribute-error
    if rule is None:
      raise NotImplementedError(f"No layout inference rule defined for {op}")
    equation_system, mapping, op_hints = rule(ctx, op)
    ctx.update(mapping)
    nonlocal global_equation_system
    global_equation_system &= equation_system
    hints.extend(op_hints)

  for op in module.body:
    inference_utils.traverse_op(op, gather_equations, pre_order=True)

  if isinstance(global_equation_system, eqns.Unsatisfiable):
    raise ValueError(
        "Failed to infer a possible set of layouts. This should only happen if "
        "user-provided layout casts are unsatisfiable."
    )

  if not enable_smem_inference:
    global_equation_system, ctx = _drop_smem(global_equation_system, ctx)

  propagation_hints, constraints = derive_hints_and_constraints(ctx.operand_and_results_for_variable)
  hints = reduce_hints(hints + propagation_hints, global_equation_system.assignments)  # pytype: disable=attribute-error
  global_equation_system &= eqns.EquationSystem(constraints=constraints)
  assert not isinstance(global_equation_system, eqns.Unsatisfiable)

  # Add additional (redundant) constraints which helps the search converge
  # faster.
  global_equation_system = eqns.saturate_distinct_from_splat(global_equation_system)
  assert not isinstance(global_equation_system, eqns.Unsatisfiable)
  global_equation_system = eqns.saturate_divides_constraints_for_equal_vars(global_equation_system)

  # Attempt to find assignments that satisfy the equation system.
  solution = find_assignments_for(
      ctx.operand_and_results_for_variable.keys(), global_equation_system, hints
  )

  if isinstance(solution, eqns.Unsatisfiable):
    raise ValueError(
        "Failed to infer a possible set of layouts. This should only happen if "
        "user-provided layout casts are unsatisfiable."
    )

  layout_for_operand_or_result = {
      k: solution[v]
      for v, ks in ctx.operand_and_results_for_variable.items()
      for k in ks
  }

  # Assigns the layouts that we found to the ops.
  assign_layouts(layout_for_operand_or_result, enable_smem_inference)

  # Sanity check: ensure that all ops have the right number of in/out layouts.
  for op in module.body:
    inference_utils.traverse_op(
        op,
        partial(
            _ensure_all_layouts_are_set,
            enable_smem_inference=enable_smem_inference,
        ),
    )



# Copyright 2025 The JAX Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Transform inference pass for the MLIR Mosaic GPU dialect.

The transform inference pass is meant to run on IR that has already been
annotated with layouts (see `layout_inference.py` for the relevant pass).
"""

from collections.abc import Callable
from functools import partial
import math
from typing import cast

from jax._src.lib import mosaic_gpu_dialect as mgpu
from jax._src.lib.mlir import ir
from jax._src.lib.mlir.dialects import arith
from jax._src.lib.mlir.dialects import gpu
from jax._src.lib.mlir.dialects import llvm
from jax._src.lib.mlir.dialects import memref
from jax._src.lib.mlir.dialects import vector

from . import inference_utils
from . import layouts as layouts_lib
from . import utils


OptionalTransforms = tuple[list[ir.Attribute], list[ir.Attribute]] | None
TransformInferenceRule = Callable[[ir.OpView], OptionalTransforms]
_transform_inference_rules: dict[str, TransformInferenceRule] = {}


def _add_transform_inference_rule(
    op: type[ir.OpView], rule: TransformInferenceRule
):
  if op is not None:
    _transform_inference_rules[op.OPERATION_NAME] = rule  # pytype: disable=attribute-error
  return rule


def _set_transform_attributes(
    op: ir.OpView,
    in_transforms: list[ir.Attribute],
    out_transforms: list[ir.Attribute],
):
  op.attributes["in_transforms"] = ir.ArrayAttr.get(in_transforms)
  op.attributes["out_transforms"] = ir.ArrayAttr.get(out_transforms)


def _resolve_transforms(
    transforms: ir.ArrayAttr | None,
    other_transforms: ir.ArrayAttr | None,
) -> ir.ArrayAttr | None:
  """Resolves two sets of competing transforms to a single compatible set.

  Args:
    transforms: one optional set of transforms.
    other_transforms: another optional set of transforms.

  Returns:
    A single set of transforms that is compatible with both `transforms` and
    `other_transforms`, or `None` if both transforms are `None`.
  Raises:
    NotImplementedError: if the two sets of transforms can't be resolved to a
      single set.
  """
  if transforms is None:
    return other_transforms

  if other_transforms is None:
    return transforms

  if len(transforms) != len(other_transforms):
    raise NotImplementedError(
        f"Conflicting transforms {transforms} != {other_transforms}."
    )

  new_transforms = []
  for a, b in zip(transforms, other_transforms, strict=True):
    if a == b:
      new_transforms.append(a)
    elif mgpu.TileTransformAttr.isinstance(a) and mgpu.TileTransformAttr.isinstance(b):
      a = mgpu.TileTransformAttr(a)
      b = mgpu.TileTransformAttr(b)
      if len(a.tiling) != len(b.tiling):
        raise ValueError(f"Conflicting tile transforms {a} != {b}.")
      new_tiling = []
      for tile_a, tile_b in zip(a.tiling, b.tiling):
        new_tiling.append(math.gcd(tile_a, tile_b))
      new_transforms.append(mgpu.TileTransformAttr.get(new_tiling))
    else:
      raise NotImplementedError(f"Unsupported transforms {a} and {b}")

  return ir.ArrayAttr.get(new_transforms)


def _transforms_from_uses(op: ir.OpView) -> ir.ArrayAttr | None:
  transforms = None

  for result_use in cast(ir.OpResult, op.result).uses:
    consumer = result_use.owner
    op_user = consumer.operands[result_use.operand_number]
    user_transforms = inference_utils.in_transforms_for_operand(
        consumer, op_user
    )
    transforms = _resolve_transforms(transforms, user_transforms)
  return transforms


def _infer_transforms_for_mma_ref(
    ref_ty: ir.MemRefType, max_swizzle: mgpu.SwizzlingMode
) -> tuple[ir.ArrayAttr, mgpu.SwizzlingMode]:
  if len(ref_ty.shape) != 2:
    raise ValueError(f"Expected a 2D memref, got {ref_ty}")

  element_bytewidth = utils.bytewidth(ref_ty.element_type)
  transposed = utils.is_memref_transposed(ref_ty)
  minor_dim = ref_ty.shape[0 if transposed else 1]
  major_tiling = 8

  # Try tiling with all swizzling modes starting from the largest one.
  for swizzle in [
      mgpu.SwizzlingMode.k128ByteSwizzle,
      mgpu.SwizzlingMode.k64ByteSwizzle,
      mgpu.SwizzlingMode.k32ByteSwizzle,
      mgpu.SwizzlingMode.kNoSwizzle,
  ]:
    if swizzle > max_swizzle:
      continue
    swizzle_elems = swizzle // element_bytewidth
    if minor_dim % swizzle_elems == 0:
      minor_tiling = swizzle_elems
      inferred_swizzle = swizzle
      break
  else:
    # No valid tile transform can be inferred.
    raise ValueError(f"{ref_ty.shape} is not a valid WGMMA shape")

  if transposed:
    tiling = (minor_tiling, major_tiling)
  else:
    tiling = (major_tiling, minor_tiling)
  return (
      ir.ArrayAttr.get([
          mgpu.TileTransformAttr.get(tiling),
          mgpu.SwizzleTransformAttr.get(minor_tiling * element_bytewidth),
      ]),
      inferred_swizzle,
  )


@partial(_add_transform_inference_rule, mgpu.WGMMAOp)
def infer_wgmma_transforms(op: mgpu.WGMMAOp) -> OptionalTransforms:
  b_transforms, b_swizzle = _infer_transforms_for_mma_ref(
      ir.MemRefType(op.b.type), max_swizzle=mgpu.SwizzlingMode.k128ByteSwizzle
  )
  if not ir.MemRefType.isinstance(op.a.type):
    return [b_transforms], []

  a_transforms, a_swizzle = _infer_transforms_for_mma_ref(
      ir.MemRefType(op.a.type), max_swizzle=b_swizzle
  )
  if a_swizzle != b_swizzle:
    # The swizzle for a and b has to match.
    b_transforms, b_swizzle = _infer_transforms_for_mma_ref(
        ir.MemRefType(op.b.type), max_swizzle=a_swizzle
    )
    assert a_swizzle == b_swizzle
  return [a_transforms, b_transforms], []


@partial(_add_transform_inference_rule, mgpu.TcGen05MMAOp)
def infer_tcgen05_mma_transforms(op: mgpu.TcGen05MMAOp) -> OptionalTransforms:
  b_transforms, _ = _infer_transforms_for_mma_ref(
      ir.MemRefType(op.b.type), max_swizzle=mgpu.SwizzlingMode.k128ByteSwizzle
  )
  if not utils.is_smem_ref(op.a.type):
    return [b_transforms], []

  a_transforms, _ = _infer_transforms_for_mma_ref(
      ir.MemRefType(op.a.type), max_swizzle=mgpu.SwizzlingMode.k128ByteSwizzle
  )
  return [a_transforms, b_transforms], []


@partial(_add_transform_inference_rule, mgpu.AsyncStoreOp)
def _infer_async_store_transforms(op: mgpu.AsyncStoreOp) -> OptionalTransforms:
  in_transforms = inference_utils.value_transforms(op.source)
  return None if in_transforms is None else ([in_transforms], [])


@partial(_add_transform_inference_rule, mgpu.AsyncLoadOp)
def _infer_async_load_transforms(op: mgpu.AsyncLoadOp) -> OptionalTransforms:
  in_transforms = inference_utils.value_transforms(op.destination)
  return None if in_transforms is None else ([in_transforms], [])


@partial(_add_transform_inference_rule, vector.LoadOp)
@partial(_add_transform_inference_rule, vector.StoreOp)
def _infer_vector_load_store_transforms(
    op: vector.LoadOp | vector.StoreOp,
) -> OptionalTransforms:
  for i in op.indices:
    index_defining_op = i.owner.opview
    if (
        not isinstance(index_defining_op, arith.ConstantOp)
        or index_defining_op.literal_value != 0
    ):
      # TODO(bchetioui): handle slicing.
      raise NotImplementedError(
          f"Only constants with value 0 are supported as indices for {op}"
      )

  if isinstance(op, vector.LoadOp):
    [layout_attr] = inference_utils.out_layouts(op)
  else:
    assert isinstance(op, vector.StoreOp)
    [layout_attr] = inference_utils.in_layouts(op)

  layout = layouts_lib.from_layout_attr(layout_attr)
  if inference_utils.is_mma_layout(layout):
    base_type = ir.MemRefType(op.base.type)
    layout_transforms, _ = _infer_transforms_for_mma_ref(
        base_type,
        max_swizzle=mgpu.SwizzlingMode.k128ByteSwizzle,
    )
  else:
    layout_transforms = None

  transforms = inference_utils.value_transforms(op.base)
  transforms = _resolve_transforms(transforms, layout_transforms)
  return None if transforms is None else ([transforms], [])


@partial(_add_transform_inference_rule, memref.StoreOp)
def _infer_memref_store_transforms(op: memref.StoreOp) -> OptionalTransforms:
  # memref.store is only used for scalar operations, so there are no transforms.
  ref_shape = ir.MemRefType(op.memref.type).shape
  if ref_shape != [] and ref_shape != [1]:
    raise NotImplementedError(
        f"Only scalar memrefs are supported, got {ref_shape}"
    )

  return None

@partial(_add_transform_inference_rule, mgpu.SliceSMEMOp)
def _infer_slice_smem_transforms(op: mgpu.SliceSMEMOp) -> OptionalTransforms:
  transforms = _transforms_from_uses(op)
  return None if transforms is None else ([], [transforms])


@partial(_add_transform_inference_rule, memref.ViewOp)
def _infer_memref_view_transforms(op: memref.ViewOp) -> OptionalTransforms:
  if not isinstance(op.source.owner.opview, gpu.DynamicSharedMemoryOp):
    raise NotImplementedError(
        "Memref view transforms are only inferred when the op is a direct user "
        f"of a DynamicSharedMemoryOp but got {op}."
    )
  transforms = inference_utils.value_transforms(op.source)
  if transforms is not None:
    raise NotImplementedError(
        "memref view with in_transforms aren't yet supported"
    )
  transforms = _transforms_from_uses(op)

  # TODO(bchetioui): do we actually need to assign a transform to the input of
  # the view op? Presumably, it'll only be used to access scratch memory.
  return None if transforms is None else ([], [transforms])


@partial(_add_transform_inference_rule, llvm.UndefOp)
def _infer_llvm_undef_transforms(op: llvm.UndefOp) -> OptionalTransforms:
  # This is only needed for tests and will be deleted once we move to SMEM
  # inference in the new constraint-based system.
  transforms = _transforms_from_uses(op)

  if transforms is None:
    return None
  return [], [transforms]


def _get_tile_and_swizzle_transforms(
    transforms: ir.ArrayAttr,
) -> tuple[ir.Attribute, ir.Attribute]:
  if len(transforms) == 2:
    tile_transform, swizzle_transform = transforms
    if not (
        mgpu.TileTransformAttr.isinstance(tile_transform)
        and mgpu.SwizzleTransformAttr.isinstance(swizzle_transform)
    ):
      raise NotImplementedError(f"Unsupported transforms {transforms}.")
    return tile_transform, swizzle_transform
  else:
    raise NotImplementedError(f"Unsupported transforms {transforms}.")


# This is used by Pallas' "_handle_indexing" memory transform.
@partial(_add_transform_inference_rule, memref.SubViewOp)
def _infer_memref_subview_transforms(
    op: memref.SubViewOp,
) -> OptionalTransforms:
  transforms = _transforms_from_uses(op)
  in_transforms = inference_utils.value_transforms(op.source)
  transforms = _resolve_transforms(transforms, in_transforms)

  if transforms is None:
    return None
  if not transforms:
    empty = ir.ArrayAttr.get([])
    return [empty], [empty]

  # Here, we have some transforms to propagate one way or the other. For now,
  # we implement only the following basic propagation rules:
  #  - A tile transform can be propagated bidirectionally if the axes being
  #    tiled are not sliced, and are the logical minor axes of the source.
  #  - A swizzle transform can be propagated towards the input of a subview if
  #    the physical minormost dimension is unchanged.
  #  - We only propagate transforms if they consist of a single tile transform
  #    and a single swizzle transform.
  # TODO(bchetioui): implement more complex propagation rules.
  tile_transform, swizzle_transform = _get_tile_and_swizzle_transforms(transforms)

  # Check swizzle transform propagation.
  strides, _ = ir.MemRefType.get_strides_and_offset(op.source.type)
  minor_dim = strides.index(min(strides))
  if op.source.type.shape[minor_dim] != op.static_sizes[minor_dim]:
    raise NotImplementedError(
        "Swizzle transforms can only propagated if the minor dimension is "
        "unchanged."
    )

  is_dynamic = lambda x: ir.ShapedType.is_dynamic_size(x)

  # Check tile transform propagation.
  old_tiling = mgpu.TileTransformAttr(tile_transform).tiling
  num_tiled_axes = len(old_tiling)
  last_n_dims = op.source.type.shape[-num_tiled_axes:]
  last_n_sizes = list(op.static_sizes)[-num_tiled_axes:]
  last_n_offsets = list(op.static_offsets)[-num_tiled_axes:]

  if any(is_dynamic(x) for x in last_n_sizes):
    raise NotImplementedError(
        "Subview transforms with dynamic sizes are not supported."
    )

  num_non_tiled_axes = len(op.source.type.shape) - num_tiled_axes
  non_tiled_offsets = list(op.static_offsets)[:num_non_tiled_axes]
  dynamic_index = sum(1 for x in non_tiled_offsets if is_dynamic(x))

  for i in range(len(last_n_offsets)):
    if is_dynamic(last_n_offsets[i]):
      if utils.is_known_divisible(
          op.offsets[dynamic_index], last_n_sizes[i]
      ):
        last_n_offsets[i] = last_n_sizes[i]
      else:
        # This will force a tiling of 1 along this axis. This is a safe choice
        # (since we couldn't infer a better one) but might not be optimal.
        last_n_offsets[i] = 1
      dynamic_index += 1

  new_tiling = [
      math.gcd(*xs)
      for xs in zip(
          last_n_sizes, last_n_dims, last_n_offsets, old_tiling, strict=True
      )
  ]

  new_transforms = ir.ArrayAttr.get(
      [mgpu.TileTransformAttr.get(new_tiling), swizzle_transform]
  )
  return [new_transforms], [new_transforms]


@partial(_add_transform_inference_rule, memref.TransposeOp)
def _infer_memref_transpose_transforms(
    op: memref.TransposeOp,
) -> OptionalTransforms:
  in_ty = ir.MemRefType(op.in_.type)
  if len(in_ty.shape) != 2:
    raise NotImplementedError(f"Only 2D memrefs are supported, got {in_ty}")
  in_strides, _ = in_ty.get_strides_and_offset()
  out_strides, _ = ir.MemRefType(op.result.type).get_strides_and_offset()
  transpose = in_strides != out_strides

  out_transforms = _transforms_from_uses(op)
  in_transforms: list[ir.Attribute] = []
  if not transpose:
    if out_transforms:
      in_transforms.extend(*out_transforms)
  else:
    tile_transform, swizzle_transform = _get_tile_and_swizzle_transforms(
        out_transforms
    )
    transposed_tiling = mgpu.TileTransformAttr(tile_transform).tiling[::-1]
    in_transforms.append(mgpu.TileTransformAttr.get(transposed_tiling))
    in_transforms.append(swizzle_transform)

  return [ir.ArrayAttr.get(in_transforms)], [out_transforms]


# `memref.load` is used to load barrier phases---the rule needn't do anything
# interesting, but we need to have it in order to avoid crashing on it.
@partial(_add_transform_inference_rule, memref.LoadOp)
def _infer_memref_load_transforms(op: memref.LoadOp) -> OptionalTransforms:
  if not ir.MemRefType(op.memref.type).shape:
    # memref.load returns a scalar, so there is nothing interesting to do here.
    return None
  raise NotImplementedError("Non-scalar memref.load transforms")


@partial(_add_transform_inference_rule, memref.CastOp)
def _infer_memref_cast_transforms(
    op: memref.CastOp,
) -> OptionalTransforms:
  transforms = _transforms_from_uses(op)
  in_transforms = inference_utils.value_transforms(op.source)
  transforms = _resolve_transforms(transforms, in_transforms)
  if transforms is None:
    return None
  return [transforms], [transforms]


@partial(_add_transform_inference_rule, mgpu.WithTransformsOp)
def _infer_mgpu_with_transforms_transforms(
    op: mgpu.WithTransformsOp,
) -> OptionalTransforms:
  # Do not change the manually provided transforms.
  return [op.transforms], [op.transforms]


@partial(_add_transform_inference_rule, mgpu.TmemAllocOp)
def _infer_tmem_alloc_transforms(op: mgpu.TmemAllocOp) -> OptionalTransforms:
  del op
  return [], []


@partial(_add_transform_inference_rule, mgpu.CustomPrimitiveOp)
def _infer_mgpu_custom_primitive_transforms(
    op: mgpu.CustomPrimitiveOp,
) -> OptionalTransforms:
  # Do not change the manually provided transforms.
  return list(op.in_transforms), []


def infer_transforms(module: ir.Module):
  """Infers transforms for the given module.

  Transforms are to memrefs what layouts are to vectors. More specifically,
  transforms describe mappings between SMEM refs and GMEM refs, and are
  determined based on how SMEM refs are used. For that reason, we always
  annotate and apply memrefs on SMEM refs.

  The pass is meant to be called on a module where layouts have been fully
  specified. We error out if two distinct sets of transforms are competing to
  annotate the same memref.
  """
  def inference_step(op: ir.Operation):
    if not inference_utils.should_have_transforms(op):
      return
    elif inference_rule := _transform_inference_rules.get(op.OPERATION_NAME, None):  # pytype: disable=attribute-error
      pass
    else:
      raise NotImplementedError(f"Can not infer transforms for {op}")

    maybe_transforms = inference_rule(op)
    if maybe_transforms is None:
      return

    _set_transform_attributes(op, *maybe_transforms)

  # We alternate a few backwards propagation (starting from vector users), and
  # forward propagation (to feed into the async loads and stores) passes in
  # order to enable more complex inference situations.
  #
  # TODO(bchetioui): Replace this with a more generic inference.
  inference_passes = [
      inference_utils.TraversalOrder.BACKWARDS,
      inference_utils.TraversalOrder.FORWARD,
      inference_utils.TraversalOrder.BACKWARDS,
      inference_utils.TraversalOrder.FORWARD,
  ]
  for traversal_order in inference_passes:
    for op in module.body:
      inference_utils.traverse_op(op, inference_step, traversal_order)

  # All ops that should have transforms but have no transforms inferred so far
  # are assigned an empty sets of transforms. E.g., this happens in kernels with
  # only pointwise operations.
  def set_empty_transforms(op: ir.Operation):
    if (
        inference_utils.should_have_transforms(op)
        and not inference_utils.has_in_transforms_set(op)
        and not inference_utils.has_out_transforms_set(op)
    ):
      ins = [
          ir.ArrayAttr.get([])
          for o in op.operands
          if inference_utils.is_transformable_smem_memref(o)
      ]
      outs = [
          ir.ArrayAttr.get([])
          for r in op.results
          if inference_utils.is_transformable_smem_memref(r)
      ]
      _set_transform_attributes(op, ins, outs)

  for op in module.body:
    inference_utils.traverse_op(
        op, set_empty_transforms, inference_utils.TraversalOrder.FORWARD
    )



# Copyright 2024 The JAX Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Colocated Python function API implementation."""

from __future__ import annotations

import dataclasses
import inspect
import random
import threading
from typing import Any
from collections.abc import Callable, Sequence

import jax
from jax._src import api
from jax._src import tree_util
from jax._src import util
from jax._src.interpreters import pxla
from jax._src.lib import xla_client as xc
from jax._src.traceback_util import api_boundary
from jax._src.util import wraps
from jax.experimental.colocated_python import func_backend
from jax.experimental.colocated_python.serialization import _deserialize_specs, _make_specs_for_serialized_specs, _serialize, _serialize_specs
from jax.extend.ifrt_programs import ifrt_programs

ShapeDtypeStructTree = Any  # PyTree[api.ShapeDtypeStruct]


@dataclasses.dataclass(frozen=True, slots=True)
class FunctionInfo:
  """User function wrapped by colocated_python."""

  fun: Callable[..., Any]
  fun_sourceinfo: str | None
  fun_signature: inspect.Signature | None


@dataclasses.dataclass(frozen=True, slots=True)
class Specialization:
  """Specialization for a colocated_python function."""

  in_specs_treedef: tree_util.PyTreeDef | None = None
  in_specs_leaves: tuple[api.ShapeDtypeStruct, ...] | None = None
  out_specs_fn: Callable[..., ShapeDtypeStructTree] | None = None
  out_specs_treedef: tree_util.PyTreeDef | None = None
  out_specs_leaves: tuple[api.ShapeDtypeStruct, ...] | None = None
  devices: xc.DeviceList | None = None

  def update(
      self,
      *,
      in_specs_treedef: tree_util.PyTreeDef | None = None,
      in_specs_leaves: tuple[api.ShapeDtypeStruct, ...] | None = None,
      out_specs_fn: Callable[..., ShapeDtypeStructTree] | None = None,
      out_specs_treedef: tree_util.PyTreeDef | None = None,
      out_specs_leaves: tuple[api.ShapeDtypeStruct, ...] | None = None,
      devices: Sequence[jax.Device] | xc.DeviceList | None = None,
  ):
    """Creates a new specialization with overrides."""
    if in_specs_treedef is None:
      in_specs_treedef = self.in_specs_treedef
    elif self.in_specs_treedef is not None:
      raise ValueError("in_specs already specified")
    if in_specs_leaves is None:
      in_specs_leaves = self.in_specs_leaves
    elif self.in_specs_leaves is not None:
      raise ValueError("in_specs already specified")

    if out_specs_fn is None:
      out_specs_fn = self.out_specs_fn
    elif self.out_specs_fn is not None:
      raise ValueError("out_specs_fn already specified")

    if out_specs_treedef is None:
      out_specs_treedef = self.out_specs_treedef
    elif self.out_specs_treedef is not None:
      raise ValueError("out_specs already specified")
    if out_specs_leaves is None:
      out_specs_leaves = self.out_specs_leaves
    elif self.out_specs_leaves is not None:
      raise ValueError("out_specs already specified")

    if devices is None:
      devices = self.devices
    elif self.devices is not None:
      raise ValueError("devices already specified")
    elif not isinstance(devices, xc.DeviceList):
      devices = xc.DeviceList(tuple(devices))

    return Specialization(
        in_specs_treedef,
        in_specs_leaves,
        out_specs_fn,
        out_specs_treedef,
        out_specs_leaves,
        devices,
    )


def _get_spec(x: Any) -> api.ShapeDtypeStruct:
  """Extracts a spec for a value, which must be a JAX Array."""
  # TODO(hyeontaek): Allow Python values and automatically apply `shard_arg`
  # with a suitable sharding and layout.
  if not isinstance(x, jax.Array):
    raise ValueError(
        "colocated_python only supports jax.Array as input and output, but got"
        f" {type(x)}."
    )
  return api.ShapeDtypeStruct(shape=x.shape, dtype=x.dtype, sharding=x.sharding)


def _infer_devices_from_args(args: Sequence[Any]) -> xc.DeviceList | None:
  """Returns a representative device list from function call arguments."""
  device_list_set: set[xc.DeviceList] = set()
  for x in args:
    sharding = getattr(x, "sharding", None)
    if sharding is not None:
      device_list_set.add(x.sharding._internal_device_list)
  if not device_list_set:
    return None
  if len(device_list_set) != 1:
    raise ValueError(
        "All arguments must use the same device list, but got"
        f" multiple device lists: {device_list_set}."
    )
  return device_list_set.pop()


def _compile_to_executable(
    name: str,
    fun: Callable[..., Any],
    in_specs_treedef: tree_util.PyTreeDef,
    in_specs_leaves: tuple[api.ShapeDtypeStruct, ...],
    out_specs_treedef: tree_util.PyTreeDef,
    out_specs_leaves: tuple[api.ShapeDtypeStruct, ...],
    devices: xc.DeviceList,
) -> Callable[..., Any]:
  """Compiles a Python function into a runtime executable."""
  fun_and_specialization = (
      fun,
      in_specs_treedef,
      in_specs_leaves,
      out_specs_treedef,
      out_specs_leaves,
      devices,
  )
  pickled_function = _serialize(fun_and_specialization)
  program = ifrt_programs.make_colocated_python_program(
      name, pickled_function, devices, in_specs_leaves, out_specs_leaves
  )
  ifrt_client = devices[0].client
  out_sdss = tuple(
      jax.core.ShapedArray(sds.shape, sds.dtype) for sds in out_specs_leaves
  )
  out_shardings = tuple(sds.sharding for sds in out_specs_leaves)
  try:
    compile_options = ifrt_programs.make_colocated_python_compile_options()
    loaded_executable = ifrt_client.compile_ifrt_program(
        program, compile_options
    )
    out_handlers = pxla.global_avals_to_results_handler(
        out_sdss, out_shardings, committed=True  # type: ignore
    ).handlers

    def call(*args, **kwargs):
      args_leaves = tree_util.tree_leaves((args, kwargs))
      execute_result = loaded_executable.execute_sharded(
          args_leaves, with_tokens=False
      )
      results = execute_result.consume_with_handlers(out_handlers)
      return tree_util.tree_unflatten(out_specs_treedef, results)

    return call
  except jax.errors.JaxRuntimeError as e:
    # TODO(hyeontaek): Implement colocated Python support in McJAX and remove
    # this fallback path.
    if "PjRtCompiler requires an HloProgram" in str(e):
      return fun
    raise


def _make_output_specs_and_push_result_fun(
    info: FunctionInfo, specialization: Specialization, uid: int
) -> Callable[..., Any]:
  """Creates a function that computes output specs and pushes the result to the result store."""
  assert specialization.in_specs_treedef is not None
  assert specialization.in_specs_leaves is not None
  assert specialization.out_specs_treedef is None
  assert specialization.out_specs_leaves is None
  assert specialization.devices is not None

  devices = specialization.devices

  def lowered_fun(*args, **kwargs) -> jax.Array:
    result = info.fun(*args, **kwargs)
    result_leaves, out_treedef = tree_util.tree_flatten(result)
    out_spec_leaves = tuple(_get_spec(x) for x in result_leaves)
    func_backend.SINGLETON_RESULT_STORE.push(uid, result_leaves)
    return _serialize_specs(out_treedef, out_spec_leaves, devices)

  out_specs_leaves, out_specs_treedef = tree_util.tree_flatten(
      _make_specs_for_serialized_specs(specialization.devices),
  )
  name = getattr(info.fun, "__name__", "unknown")
  name = f"{name}_output_specs_and_push_result"
  return _compile_to_executable(
      name=name,
      fun=lowered_fun,
      in_specs_treedef=specialization.in_specs_treedef,
      in_specs_leaves=specialization.in_specs_leaves,
      out_specs_treedef=out_specs_treedef,
      out_specs_leaves=tuple(out_specs_leaves),
      devices=specialization.devices,
  )


def _make_pop_result_fun(
    info: FunctionInfo, specialization: Specialization, uid: int
) -> Callable[..., Any]:
  """Makes a function that pops results from the result store."""
  assert specialization.out_specs_treedef is not None
  assert specialization.out_specs_leaves is not None
  assert specialization.devices is not None

  out_specs_treedef = specialization.out_specs_treedef

  def lowered_fun():
    result_leaves = func_backend.SINGLETON_RESULT_STORE.pop(uid)
    return tree_util.tree_unflatten(out_specs_treedef, result_leaves)

  in_specs_leaves, in_specs_treedef = tree_util.tree_flatten((
      # args
      (),
      # kwargs
      {},
  ))
  name = getattr(info.fun, "__name__", "unknown")
  name = f"{name}_pop_result"
  return _compile_to_executable(
      name=name,
      fun=lowered_fun,
      in_specs_treedef=in_specs_treedef,
      in_specs_leaves=tuple(in_specs_leaves),
      out_specs_treedef=specialization.out_specs_treedef,
      out_specs_leaves=specialization.out_specs_leaves,
      devices=specialization.devices,
  )


def _make_async_execution_fun(
    info: FunctionInfo, specialization: Specialization
) -> Callable[..., Any]:
  """Makes a function that asynchronously executes the function."""
  assert specialization.in_specs_treedef is not None
  assert specialization.in_specs_leaves is not None
  assert specialization.out_specs_treedef is not None
  assert specialization.out_specs_leaves is not None
  assert specialization.devices is not None

  name = getattr(info.fun, "__name__", "unknown")
  return _compile_to_executable(
      name=name,
      fun=info.fun,
      in_specs_treedef=specialization.in_specs_treedef,
      in_specs_leaves=specialization.in_specs_leaves,
      out_specs_treedef=specialization.out_specs_treedef,
      out_specs_leaves=specialization.out_specs_leaves,
      devices=specialization.devices,
  )


@jax._src.util.cache(max_size=None)
def _get_specialized_func(
    info: FunctionInfo, specialization: Specialization
) -> Callable[..., Any]:
  """Returns a specialized function for the given specialization."""
  util.test_event("colocated_python_func._get_specialized_func")
  assert specialization.in_specs_treedef is not None
  assert specialization.in_specs_leaves is not None
  assert specialization.devices is not None
  uid = random.getrandbits(63)

  mutex = threading.Lock()
  # Asynchronous execution function that has known output_specs.
  async_execution_func = None

  def specialized_func(*args, **kwargs):
    """Specialized function to be executed with given args and kwargs."""
    nonlocal specialization, async_execution_func
    with mutex:
      if async_execution_func is None:
        if specialization.out_specs_treedef is None:
          if specialization.out_specs_fn is None:
            serialized_out_specs = _make_output_specs_and_push_result_fun(
                info, specialization, uid
            )(*args, **kwargs)

            # Waits for the output_specs. This may block.
            out_specs_treedef, out_specs_leaves = _deserialize_specs(
                serialized_out_specs
            )

            # Subsequent calls would use async_execution_func with discovered
            # output_specs.
            specialization = specialization.update(
                out_specs_treedef=out_specs_treedef,
                out_specs_leaves=out_specs_leaves,
            )
            async_execution_func = _make_async_execution_fun(
                info, specialization
            )

            return _make_pop_result_fun(info, specialization, uid)()
          else:
            # Compute out_specs using out_specs_fn and inputs.
            args_specs, kwargs_specs = tree_util.tree_map(
                _get_spec, (args, kwargs)
            )
            out_specs = specialization.out_specs_fn(*args_specs, **kwargs_specs)
            # Type checking is ignored to silence mypy error: Incompatible types
            # in assignment (expression has type "list[Any]", variable has type
            # "tuple[ShapeDtypeStruct, ...]")  [assignment]
            out_specs_leaves, out_specs_treedef = tree_util.tree_flatten(  # type: ignore[assignment]
                out_specs
            )
            specialization = specialization.update(
                out_specs_treedef=out_specs_treedef,
                out_specs_leaves=tuple(out_specs_leaves),
            )
            async_execution_func = _make_async_execution_fun(
                info, specialization
            )
            # Fall-through.
        else:
          async_execution_func = _make_async_execution_fun(info, specialization)
          # Fall-through.

    # Asynchronous execution runs outside of the mutex to allow concurrent
    # execution for inline executors.
    return async_execution_func(*args, **kwargs)

  return specialized_func


def make_callable(
    fun: Callable[..., Any],
    fun_sourceinfo: str | None,
    fun_signature: inspect.Signature | None,
):
  """Makes a colocated Python callable."""
  return _make_callable(
      FunctionInfo(fun, fun_sourceinfo, fun_signature), Specialization()
  )


def _make_callable(info: FunctionInfo, specialization: Specialization):
  """Internal implementation of make_callable."""

  def specialize(
      in_specs: ShapeDtypeStructTree | None = None,
      out_specs_fn: Callable[..., ShapeDtypeStructTree] | None = None,
      devices: Sequence[jax.Device] | None = None,
  ):
    """Returns a colocated Python callable with extra specialization.

    Args:
      in_specs: Optionally specifies the expected input specs. Input specs are
        expressed as a `PyTree[ShapeDtypeStruct]` for `(args, kwargs)` of a
        function call.
      out_specs_fn: Optionally specifies a function that computes the output
        specs from input specs. If unspecified, colocated Python will compute
        the output specs during the very first execution, and this execution
        will be synchronous.
      devices: Optionally specifies the devices to execute the function on. Must
        be provided if `in_specs` has no leaves because devices cannot be
        inferred from input specs or arguments.

    Returns:
      A colocated Python callable with extra specialization.
    """
    # TODO(hyeontaek): Allow unspecified devices for zero-leaf `in_specs` if
    # `out_specs_fn(in_specs)` returns at least one leaf that we can use for
    # inferring `devices`.
    if in_specs is None:
      in_specs_leaves, in_specs_treedef = None, None
    else:
      in_specs_leaves_list, in_specs_treedef = tree_util.tree_flatten(in_specs)
      in_specs_leaves = tuple(in_specs_leaves_list)
    return _make_callable(
        info,
        specialization.update(
            in_specs_treedef=in_specs_treedef,
            in_specs_leaves=in_specs_leaves,
            out_specs_fn=out_specs_fn,
            devices=devices,
        ),
    )

  @api_boundary
  def __call__(*args, **kwargs):
    """Executes the given Python function on the same devices as the arguments or as specialized.

    If the callable has not been specialized with output shapes and shardings
    (see `specialize` above), the very first call will run synchronously to
    discover output shapes and shardings, and will run asynchronously after. If
    specialized with output shapes and shardings, every execution of the
    callable will be asynchronous.
    """
    args_leaves, in_specs_treedef = tree_util.tree_flatten((args, kwargs))

    in_specs_leaves = tuple(_get_spec(x) for x in args_leaves)
    if specialization.in_specs_treedef is None:
      # Allow input polymorphism by applying input_specs specialization
      # temporarily for this call.
      return _make_callable(
          info,
          specialization.update(
              in_specs_treedef=in_specs_treedef,
              in_specs_leaves=in_specs_leaves,
          ),
      )(*args, **kwargs)

    if specialization.devices is None:
      devices = _infer_devices_from_args(args_leaves)
      if devices is None:
        raise ValueError(
            "No devices found. colocated_python function without input"
            " arguments must be first specialized with devices."
        )
      # Allow device polymorphism by applying devices specialization temporarily
      # for this call.
      return _make_callable(info, specialization.update(devices=devices))(
          *args, **kwargs
      )

    # Assertion is added to silence mypy error: Unsupported operand types for !=
    # ("PyTreeDef" and "None")  [operator]
    assert isinstance(specialization.in_specs_treedef, tree_util.PyTreeDef)

    # If input_specs is known, verify that it matches actual inputs.
    if (specialization.in_specs_treedef != in_specs_treedef
        or specialization.in_specs_leaves != in_specs_leaves):
      raise ValueError(
          "Input specs in specialization and input specs of arguments must have"
          " the same pytree structure, but they have the following structural"
          " differences:\n"
          + ("\n".join(
                f"   - {tree_util.keystr(path)} is a {thing1} in value 1 and"
                f" a {thing2} in  value 2, so {explanation}.\n"
                for path, thing1, thing2, explanation in tree_util.equality_errors_pytreedef(
                    specialization.in_specs_treedef, in_specs_treedef
                ))))

    return _get_specialized_func(info, specialization)(*args, **kwargs)

  __call__ = wraps(info.fun)(__call__)
  __call__.specialize = specialize
  return __call__



# Copyright 2023 The JAX Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Contains registrations for pallas_call on TPU."""

from __future__ import annotations

from collections.abc import Sequence
import dataclasses
import os
import tempfile
from typing import cast

import jax
from jax import dtypes
from jax._src import config
from jax._src import core as jax_core
from jax._src import frozen_dict
from jax._src import sharding_impls
from jax._src import tpu_custom_call
from jax._src.state import types as state_types
from jax._src.interpreters import mlir
from jax._src.lib.mlir import ir
from jax._src.lib.mlir import passmanager
from jax._src.pallas import core as pallas_core
from jax._src.pallas.mosaic import core as tpu_core
from jax._src.pallas.mosaic import lowering
from jax._src.pallas.mosaic import sc_lowering
from jax._src.pallas.mosaic import verification
from jax.experimental import mosaic
from jax.experimental.mosaic.dialects import tpu


def _maybe_cast_to_int(x: jax.Array | jax_core.AbstractValue):
  """Casts boolean values to integers.

  We perform this cast because Mosaic does not directly support bool values
  for Memrefs. Instead, we load bools as integers and cast them to bools
  after loading from a memref inside of the kernel.
  """
  assert isinstance(
      x, (jax.Array, jax_core.ShapedArray, jax_core.DShapedArray,
          state_types.AbstractLinVal)
  ), type(x)
  if isinstance(x, jax.Array):
    if dtypes.issubdtype(x.dtype, jax.numpy.bool_):
      return x.astype(lowering.BOOL_MEMREF_TYPE)
    return x
  else:
    if dtypes.issubdtype(x.dtype, jax.numpy.bool_):
      if isinstance(x, state_types.AbstractLinVal):
        raise NotImplementedError  # TODO(mattjj,sharadmv)
      return jax_core.ShapedArray(x.shape, lowering.BOOL_MEMREF_TYPE)
    return x

_DUMP_PROMELA_TO = config.string_flag(
    "jax_pallas_dump_promela_to",
    default=os.getenv("JAX_PALLAS_DUMP_PROMELA_TO", ""),
    help=(
        "If set, dumps a Promela model of the kernel to the specified"
        " directory. The model can verify that the kernel is free of data"
        " races, deadlocks, etc."
    ),
)


def _get_memory_space_from_aval(
    out_aval: jax_core.AbstractValue,
) -> tpu_custom_call.MemorySpace | None:
  if not isinstance(out_aval, jax_core.ShapedArray):
    raise ValueError("Memory spaces not defined for non-ShapedArrays")
  if not isinstance(out_aval, pallas_core.ShapedArrayWithMemorySpace):
    # If we are passed a regular old ShapedArray, we don't constrain the
    # memory space
    return None
  # If we are passed an aval with an explicit memory space tag, we use it
  # to constrain the memory space.
  match out_aval.memory_space:
    case None:
      return None
    case tpu_core.MemorySpace.ANY:
      return None
    case tpu_core.MemorySpace.HBM:
      return tpu_custom_call.MemorySpace.HBM
    case tpu_core.MemorySpace.VMEM:
      return tpu_custom_call.MemorySpace.VMEM
    case tpu_core.MemorySpace.SMEM:
      return tpu_custom_call.MemorySpace.SMEM
    case tpu_core.MemorySpace.SEMAPHORE:
      return tpu_custom_call.MemorySpace.SEMAPHORE_MEM
    case tpu_core.MemorySpace.HOST:
      return tpu_custom_call.MemorySpace.HOST
  return None


def _get_memory_spaces_from_avals(
    avals: Sequence[jax_core.AbstractValue],
) -> tuple[tpu_custom_call.MemorySpace | None, ...] | None:
  memory_spaces = None
  if any(
      isinstance(aval, pallas_core.ShapedArrayWithMemorySpace) for aval in avals
  ):
    memory_spaces = tuple(map(_get_memory_space_from_aval, avals))
  return memory_spaces


def _tc_lowering_rule(
    ctx: mlir.LoweringRuleContext,
    *in_nodes,
    jaxpr: jax_core.Jaxpr,
    grid_mapping: pallas_core.GridMapping,
    input_output_aliases: tuple[tuple[int, int], ...],
    debug: bool,
    mesh: pallas_core.Mesh | None,
    cost_estimate: pallas_core.CostEstimate | None,
    out_avals: tuple[jax_core.AbstractValue, ...],
    metadata: frozen_dict.FrozenDict[str, str] | None,
    mosaic_params: tpu_core.CompilerParams,
    debug_info: jax_core.DebugInfo,
):
  del mesh
  jax_mesh = None
  axis_context = ctx.module_context.axis_context
  if axis_context is not None:
    if isinstance(axis_context, sharding_impls.SPMDAxisContext):
      jax_mesh = axis_context.mesh
  mlir_ctx = mlir.JaxIrContext()
  mlir_ctx.append_dialect_registry(mlir.upstream_dialects)
  mlir_ctx.load_all_available_dialects()
  tpu.register_dialect(mlir_ctx)

  def lower_module(for_verification: bool):
    if for_verification:
      mlir_ctx.allow_unregistered_dialects = True
    with mlir_ctx, ir.Location.unknown(mlir_ctx):
      return lowering.lower_jaxpr_to_module(
          ctx,
          grid_mapping,
          jaxpr,
          dimension_semantics=mosaic_params.dimension_semantics,
          kernel_type=mosaic_params.kernel_type,
          mesh=jax_mesh,
          for_verification=for_verification,
          dynamic_shape_replacement_enabled=pallas_core.dynamic_shapes_export_enabled(),
      )

  mosaic_module = lower_module(for_verification=False)
  if debug:
    print(f"\nThe Mosaic module for pallas_call {debug_info.func_src_info}:")
    print(mosaic_module)
  num_dyn_bounds = grid_mapping.num_dynamic_grid_bounds
  input_output_aliases = tuple(
      (a[0] + num_dyn_bounds, a[1])
      for a in input_output_aliases
  )

  if promela_dump_path := _DUMP_PROMELA_TO.value:
    num_devices = 1 if jax_mesh is None else jax_mesh.devices.size
    num_cores = (
        jax.devices()[0].num_cores
        if jax_mesh is None
        else jax_mesh.devices[0].num_cores
    )
    verification_module = lower_module(for_verification=True)
    model = verification.export_promela_model(
        verification_module, num_devices, num_cores
    )
    if promela_dump_path == "stdout":
      print(f"The Promela model for pallas_call {debug_info.func_src_info}:")
      print(model)
    else:
      if promela_dump_path == "sponge":
        promela_dump_path = os.getenv("TEST_UNDECLARED_OUTPUTS_DIR", "")
        if not promela_dump_path:
          raise ValueError(
              "TEST_UNDECLARED_OUTPUTS_DIR must be set when"
              " --jax_pallas_dump_promela_to=sponge"
          )
      dump_ctx = tempfile.NamedTemporaryFile(
          mode="w",
          prefix=mlir.sanitize_name(debug_info.func_name) + "-",
          suffix=".pml",
          dir=promela_dump_path,
          delete=False,
      )
      with dump_ctx as f:
        f.write(model)

  # Replace in_avals to physical avals.
  # This step is required for mapping logical types to physical types.
  # (e.g. PRNG key -> uint32[2])
  physical_avals = [jax_core.physical_aval(aval) for aval in ctx.avals_in]
  ctx = ctx.replace(avals_in=physical_avals)

  # Booleans are loaded into the kernel as integers.
  def _maybe_cast_inputs(*args):
    args = [_maybe_cast_to_int(x) for x in args]
    return args

  kernel_in_avals = [_maybe_cast_to_int(x) for x in ctx.avals_in]
  kernel_out_avals = [_maybe_cast_to_int(x) for x in ctx.avals_out]
  cast_ctx = ctx.replace(avals_out=kernel_in_avals)
  in_nodes = mlir.lower_fun(_maybe_cast_inputs)(cast_ctx, *in_nodes)

  # Dynamic grid bounds have to go at the front.
  dynamic_grid_args, args = in_nodes[:num_dyn_bounds], in_nodes[num_dyn_bounds:]
  kernel_ctx = ctx.replace(avals_in=kernel_in_avals, avals_out=kernel_out_avals)
  output_memory_spaces = _get_memory_spaces_from_avals(out_avals)
  input_memory_spaces = None
  if any(
      isinstance(aval, pallas_core.ShapedArrayWithMemorySpace)
      for aval in ctx.avals_in
  ):
    # TODO(sharadmv): Support dynamic grid bounds.
    if num_dyn_bounds != 0:
      raise NotImplementedError(
          "Dynamic grid bounds are not supported when specifying memory spaces for inputs."
      )
    input_memory_spaces = _get_memory_spaces_from_avals(ctx.avals_in)
  if cost_estimate is not None:
    mosaic_cost_estimate = cast(
        tpu_custom_call.CostEstimate, dataclasses.asdict(cost_estimate)
    )
  else:
    mosaic_cost_estimate = None
  if input_memory_spaces is None and output_memory_spaces is not None:
    input_memory_spaces_list: list[tpu_custom_call.MemorySpace | None] = [
        None,
    ] * len(ctx.avals_in)
    for input_output_alias in input_output_aliases:
      input_memory_spaces_list[input_output_alias[0]] = output_memory_spaces[
          input_output_alias[1]
      ]
    input_memory_spaces = tuple(input_memory_spaces_list)
  if input_memory_spaces is not None:
    # Filter out the memory spaces that are not supported for input memory
    # spaces.
    input_memory_spaces = tuple(
        i
        if i
        in {  # pylint: disable=g-long-ternary
            tpu_custom_call.MemorySpace.HBM,
            tpu_custom_call.MemorySpace.VMEM,
            tpu_custom_call.MemorySpace.SMEM,
        }
        else None
        for i in input_memory_spaces
    )
  out_nodes = mosaic.lower_module_to_custom_call(
      kernel_ctx,
      *dynamic_grid_args,
      *args,
      module=mosaic_module,
      out_type=kernel_out_avals,
      kernel_name=mlir.sanitize_name(debug_info.func_name),
      cost_estimate=mosaic_cost_estimate,
      vmem_limit_bytes=mosaic_params.vmem_limit_bytes,
      flags=mosaic_params.flags,
      allow_input_fusion=mosaic_params.allow_input_fusion,
      input_output_aliases=input_output_aliases,
      serialization_format=mosaic_params.serialization_format,
      internal_scratch_in_bytes=mosaic_params.internal_scratch_in_bytes,
      collective_id=mosaic_params.collective_id,
      has_side_effects=mosaic_params.has_side_effects,
      output_memory_spaces=output_memory_spaces,
      disable_bounds_checks=mosaic_params.disable_bounds_checks,
      input_memory_spaces=input_memory_spaces,
      metadata=dict(metadata) if metadata is not None else None,
      skip_device_barrier=mosaic_params.skip_device_barrier,
      allow_collective_id_without_custom_barrier=mosaic_params.allow_collective_id_without_custom_barrier,
      shape_invariant_numerics=mosaic_params.shape_invariant_numerics,
  )
  _maybe_cast_to_bool = (
      lambda x, aval: x.astype(jax.numpy.bool_)
      if aval.dtype == jax.numpy.bool_
      else x
  )

  def _maybe_cast_outputs(*args):
    args = [_maybe_cast_to_bool(x, aval) for x, aval in zip(args, out_avals)]
    return args

  cast_ctx = ctx.replace(avals_in=kernel_out_avals)
  return mlir.lower_fun(_maybe_cast_outputs)(cast_ctx, *out_nodes)


# TODO(sharadmv,slebedev): Dedup with _tc_lowering_rule.
def _sc_lowering_rule(
    ctx: mlir.LoweringRuleContext,
    *in_nodes,
    jaxpr: jax_core.Jaxpr,
    grid_mapping: pallas_core.GridMapping,
    mesh: pallas_core.Mesh | None,
    input_output_aliases: tuple[tuple[int, int], ...],
    debug: bool,
    mosaic_params: tpu_core.CompilerParams,
    cost_estimate: pallas_core.CostEstimate | None,
    out_avals: tuple[jax_core.AbstractValue, ...],
    backend: str | None = None,
    metadata: frozen_dict.FrozenDict[str, str] | None,
    debug_info: jax_core.DebugInfo,
):
  """Lowers a pallas_call to a Mosaic SparseCore custom call."""
  del mesh, out_avals, backend
  out_shapes = grid_mapping.out_shapes

  jax_mesh = None
  axis_context = ctx.module_context.axis_context
  if axis_context is not None:
    if isinstance(axis_context, sharding_impls.SPMDAxisContext):
      jax_mesh = axis_context.mesh

  with mlir.JaxIrContext() as mlir_ctx, ir.Location.unknown(mlir_ctx):
    mlir_ctx.append_dialect_registry(mlir.upstream_dialects)
    mlir_ctx.load_all_available_dialects()
    tpu.register_dialect(mlir_ctx)
    mosaic_module = sc_lowering.lower_jaxpr_to_module(
        ctx, jaxpr, grid_mapping, mosaic_params, mesh=jax_mesh
    )
  if debug:
    pm = passmanager.PassManager.parse("builtin.module(canonicalize)", mlir_ctx)
    pm.run(mosaic_module.operation)
    print(
        "\nThe Mosaic SparseCore module for pallas_call"
        f" {debug_info.func_src_info}:"
    )
    print(mosaic_module)
  out_avals = [jax_core.ShapedArray(s.shape, s.dtype) for s in out_shapes]
  mosaic_cost_estimate = None
  if cost_estimate is not None:
    mosaic_cost_estimate = cast(
        tpu_custom_call.CostEstimate, dataclasses.asdict(cost_estimate)
    )

  def _lower_fun(*args):
    # Dynamic grid bounds have to go at the front.
    out = mosaic.as_tpu_kernel(
        mosaic_module,
        out_avals,
        kernel_name=mlir.sanitize_name(debug_info.func_name),
        cost_estimate=mosaic_cost_estimate,
        input_output_aliases=input_output_aliases,
        metadata=metadata,
        collective_id=mosaic_params.collective_id,
    )(*args)
    return out

  return mlir.lower_fun(_lower_fun, multiple_results=True)(ctx, *in_nodes)


def pallas_call_tpu_lowering_rule(
    ctx: mlir.LoweringRuleContext,
    *in_nodes,
    jaxpr: jax_core.Jaxpr,
    grid_mapping: pallas_core.GridMapping,
    mesh: pallas_core.Mesh | None,
    input_output_aliases: tuple[tuple[int, int], ...],
    debug: bool,
    interpret: bool,
    compiler_params: dict[str, pallas_core.CompilerParams],
    cost_estimate: pallas_core.CostEstimate | None,
    out_avals: tuple[jax_core.AbstractValue, ...],
    metadata: frozen_dict.FrozenDict[str, str] | None,
):
  """Lowers a pallas_call to a Mosaic TPU custom call."""
  del interpret  # Unused.

  debug_info = jaxpr.debug_info
  if debug:
    print(f"\nThe kernel jaxpr for pallas_call {debug_info.func_src_info}:")
    print(jaxpr)

  if "mosaic_tpu" in compiler_params:
    mosaic_params = cast(tpu_core.CompilerParams, compiler_params["mosaic_tpu"])
  else:
    mosaic_params = tpu_core.CompilerParams()
  match mosaic_params.kernel_type:
    case tpu_core.KernelType.TC:
      lowering_rule = _tc_lowering_rule
    case (
        tpu_core.KernelType.SC_SCALAR_SUBCORE
        | tpu_core.KernelType.SC_VECTOR_SUBCORE
    ):
      lowering_rule = _sc_lowering_rule
    case _:
      raise ValueError(f"Unsupported kernel type: {mosaic_params.kernel_type}")
  return lowering_rule(
      ctx,
      *in_nodes,
      jaxpr=jaxpr,
      grid_mapping=grid_mapping,
      input_output_aliases=input_output_aliases,
      mesh=mesh,
      debug=debug,
      cost_estimate=cost_estimate,
      out_avals=out_avals,
      metadata=metadata,
      mosaic_params=mosaic_params,
      debug_info=debug_info,
  )



# Copyright 2024 The JAX Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Layout utilities."""

import re
from typing import assert_never

from jax._src.lib import mosaic_gpu_dialect as mgpu
from jax._src.lib.mlir import ir

from . import fragmented_array as fa
from . import launch_context


_splat_fragmented_layout_attr_pattern = re.compile(
    r"^#mosaic_gpu.WGSplatFragLayout<\[(?P<shape>.*)\]>$"
)


def to_splat_fragmented_layout_attr(layout: fa.WGSplatFragLayout) -> ir.Attribute:
  """Constructs a #mosaic_gpu.WGSplatFragLayout attribute from a WGSplatFragLayout."""
  return ir.Attribute.parse(
      f"#mosaic_gpu.WGSplatFragLayout<{list(layout.shape)}>"
  )


def from_splat_fragmented_layout_attr(attr: ir.Attribute) -> fa.WGSplatFragLayout:
  """Constructs a WGSplatFragLayout from a #mosaic_gpu.WGSplatFragLayout attribute.

  Raises:
    ValueError: If the attribute is not a #mosaic_gpu.WGSplatFragLayout
      attribute.
  """
  match = _splat_fragmented_layout_attr_pattern.fullmatch(str(attr))
  if not match:
    raise ValueError(
        f"Expected a #mosaic_gpu.WGSplatFragLayout attribute, got {attr}"
    )

  return fa.WGSplatFragLayout(
      shape=tuple(int(s) for s in match.group("shape").split(","))
  )


def is_splat_fragmented_layout(attr: ir.Attribute) -> bool:
  return bool(_splat_fragmented_layout_attr_pattern.search(str(attr)))


_strided_fragmented_layout_attr_pattern = re.compile(
    r"^#mosaic_gpu.WGStridedFragLayout<\[(?P<shape>.*)\],"
    r" (?P<vector_size>\d+)>$"
)

def to_strided_fragmented_layout_attr(
    layout: fa.WGStridedFragLayout,
) -> ir.Attribute:
  """Constructs a #mosaic_gpu.WGStridedFragLayout attribute from a WGStridedFragLayout."""
  return ir.Attribute.parse(
      f"#mosaic_gpu.WGStridedFragLayout<{list(layout.shape)},"
      f" {layout.vec_size}>"
  )


def from_strided_fragmented_layout_attr(
    attr: ir.Attribute,
) -> fa.WGStridedFragLayout:
  """Constructs a WGStridedFragLayout from a #mosaic_gpu.WGStridedFragLayout attribute.

  Raises:
    ValueError: If the attribute is not a #mosaic_gpu.WGStridedFragLayout
      attribute.
  """
  match = _strided_fragmented_layout_attr_pattern.fullmatch(str(attr))
  if not match:
    raise ValueError(
        f"Expected a #mosaic_gpu.WGStridedFragLayout attribute, got {attr}"
    )

  return fa.WGStridedFragLayout(
      shape=tuple(int(s) for s in match.group("shape").split(",")),
      vec_size=int(match.group("vector_size")),
  )


def is_strided_fragmented_layout(attr: ir.Attribute) -> bool:
  return bool(_strided_fragmented_layout_attr_pattern.search(str(attr)))


_tiled_layout_attr_pattern = re.compile(
    r"^#mosaic_gpu.TiledLayout<\[(?P<tiling>.*)\],"
    r" warp_dims\s*=\s*\[(?P<warp_dims>.*)\],"
    r" lane_dims\s*=\s*\[(?P<lane_dims>.*)\],"
    r" vector_dim\s*=\s*(?P<vector_dim>[-\d]+)>$"
)


def to_tiled_layout_attr(
    layout: fa.TiledLayout,
) -> ir.Attribute:
  """Constructs a #mosaic_gpu.TiledLayout attribute from a TiledLayout."""

  def _int_or_replicated(d: int | fa.Replicated) -> str:
    if isinstance(d, fa.Replicated):
      return f"#mosaic_gpu.Replicated<times={d.times}>"
    return str(d)

  tile_str = lambda tile: "[" + ", ".join(str(d) for d in tile) + "]"
  tiling = "[" + ", ".join(tile_str(tile) for tile in layout.tiling.tiles) + "]"
  warp_dims = (
      "[" + ",".join(_int_or_replicated(d) for d in layout.warp_dims) + "]"
  )
  lane_dims = (
      "[" + ",".join(_int_or_replicated(d) for d in layout.lane_dims) + "]"
  )

  return ir.Attribute.parse(
      f"#mosaic_gpu.TiledLayout<{tiling}, warp_dims={warp_dims},"
      f" lane_dims={lane_dims}, vector_dim={layout.vector_dim}>"
  )


_list_of_lists_delimiter = re.compile(r"\]\s*,\s*\[")
_int_pattern = re.compile(r"^(?P<num>[-\d]+)(\s*:\s*\w+)?$")
_replicated_pattern = re.compile(
    r"^#mosaic_gpu.Replicated<\s*times\s*=\s*(?P<times>\d+)\s*>\s*$"
)


def from_tiled_layout_attr(
    attr: ir.Attribute,
) -> fa.TiledLayout:
  """Constructs a TiledLayout from a #mosaic_gpu.TiledLayout attribute.

  Raises:
    ValueError: If the attribute is not a #mosaic_gpu.TiledLayout
      attribute.
  """
  match = _tiled_layout_attr_pattern.fullmatch(str(attr))
  if not match:
    raise ValueError(
        f"Expected a #mosaic_gpu.TiledLayout attribute, got {attr}"
    )

  def _int_or_replicated(replicated_dim: str) -> int | fa.Replicated:
    match = _replicated_pattern.fullmatch(replicated_dim)
    if match:
      return fa.Replicated(int(match.group("times")))
    match = _int_pattern.fullmatch(replicated_dim)
    if match:
      return int(match.group("num"))
    raise ValueError(f"Unexpected format for replicated dim {replicated_dim}")

  tiling_str = match.group("tiling")
  tile_strings = []
  if len(tiling_str) > 2:
    tile_strings = _list_of_lists_delimiter.split(tiling_str[1:-1])
  tiles = tuple(tuple(map(int, ts.split(","))) for ts in tile_strings)
  return fa.TiledLayout(
      tiling=fa.Tiling(tiles),
      warp_dims=tuple(
          _int_or_replicated(s.strip())
          for s in match.group("warp_dims").split(",")
      ),
      lane_dims=tuple(
          _int_or_replicated(s.strip())
          for s in match.group("lane_dims").split(",")
      ),
      vector_dim=int(match.group("vector_dim")),
  )


def is_tiled_layout(attr: ir.Attribute) -> bool:
  return bool(_tiled_layout_attr_pattern.search(str(attr)))


def to_layout_attr(layout: fa.FragmentedLayout) -> ir.Attribute:
  """Constructs an MLIR attribute that corresponds to the given layout."""
  match layout:
    case fa.WGSplatFragLayout():
      return to_splat_fragmented_layout_attr(layout)
    case fa.WGStridedFragLayout():
      return to_strided_fragmented_layout_attr(layout)
    case fa.TiledLayout():
      return to_tiled_layout_attr(layout)
    case _:
      raise NotImplementedError(
          f"Unsupported layout for conversion to MLIR attribute: {layout}"
      )


def from_layout_attr(attr: ir.Attribute) -> fa.FragmentedLayout:
  """Constructs a layout from an MLIR attribute."""
  if is_splat_fragmented_layout(attr):
    return from_splat_fragmented_layout_attr(attr)
  elif is_strided_fragmented_layout(attr):
    return from_strided_fragmented_layout_attr(attr)
  elif is_tiled_layout(attr):
    return from_tiled_layout_attr(attr)
  else:
    raise NotImplementedError(
        f"Unsupported layout for conversion from MLIR attribute: {attr}"
    )


def splat_is_compatible_with_tiled(
    l1: fa.WGSplatFragLayout, l2: fa.TiledLayout
) -> bool:
  # A splat layout is compatible with a tiled layout up to replication if each
  # dimension in the shape of the splat layout is divisible by the corresponding
  # dimension in the base tile shape.
  s1, s2 = l1.shape, l2.base_tile_shape
  return all(d1 % d2 == 0 for d1, d2 in zip(s1, s2))


def meet_layouts(
    layout1: fa.FragmentedLayout, layout2: fa.FragmentedLayout
) -> fa.FragmentedLayout | None:
  """Returns the "meet" of two layouts that are compatible up to replication.

  The "meet" of the two layouts is the most replicated layout that is still
  less replicated than the arguments.

  This is the dual of `join_layouts`.

  Returns:
    The "meet" of the two layouts if both layouts are compatible up to
    replication.

  Raises:
    ValueError: if the two layouts are not compatible up to replication.
  """
  if layout1 == layout2:
    return layout1

  match (layout1, layout2):
    case (fa.WGSplatFragLayout(), _):
      if isinstance(layout2, fa.TiledLayout):
        if splat_is_compatible_with_tiled(layout1, layout2):
          return layout2
      elif layout1.shape == layout2.shape:
        return layout2
    case (_, fa.WGSplatFragLayout()):
      if isinstance(layout1, fa.TiledLayout):
        if splat_is_compatible_with_tiled(layout2, layout1):
          return layout1
      elif layout1.shape == layout2.shape:
        return layout1
    case (fa.TiledLayout(), fa.TiledLayout()):
      # TODO(bchetioui): handle `TiledLayout` replication.
      raise NotImplementedError("TiledLayout replication not supported yet")

  # Layouts are not compatible up to replication.
  return None

# NOTE: We say that two layouts are compatible up to replication if the two
# layouts satisfy at least one of the following conditions together:
#
# - The two layouts are equal;
# - One of the layouts is a `WGSplatFragLayout`, and
#   * The other layout is a `WGStridedFragLayout` with the same shape;
#   * The other layout is a `TiledLayout` that can be used to tile the shape
#     embedded in the `WGSplatFragLayout`.
#
# If any of these conditions hold, then we are always able to substitute one
# layout with the other without having to reorder any data in the underlying
# array---i.e. a relayout is free.
#
# Note that there are other combinations of layouts for which relayout is free,
# but we voluntarily narrowed down our definition to span a small, useful
# subset.

def join_layouts(
    layout1: fa.FragmentedLayout, layout2: fa.FragmentedLayout
) -> fa.FragmentedLayout | None:
  """Returns the "join" of two layouts that are compatible up to replication.

  The "join" of the two layouts is the least replicated layout that is still
  more replicated than the arguments.

  This is the dual of `meet_layouts`.

  Returns:
    The "join" of the two layouts if both layouts are compatible up to
    replication.

  Raises:
    ValueError: if the two layouts are not compatible up to replication.
  """
  if layout1 == layout2:
    return layout1

  match (layout1, layout2):
    case (fa.WGSplatFragLayout(), _):
      if isinstance(layout2, fa.TiledLayout):
        if splat_is_compatible_with_tiled(layout1, layout2):
          return layout1
      elif layout1.shape == layout2.shape:
        return layout1
    case (_, fa.WGSplatFragLayout()):
      if isinstance(layout1, fa.TiledLayout):
        if splat_is_compatible_with_tiled(layout2, layout1):
          return layout2
      elif layout1.shape == layout2.shape:
        return layout2
    case (fa.TiledLayout(), fa.TiledLayout()):
      # TODO(bchetioui): handle `TiledLayout` replication.
      raise NotImplementedError("TiledLayout replication not supported yet")

  # Layouts are not compatible up to replication.
  return None


def has_any_replication(layout: fa.FragmentedLayout) -> bool:
  match layout:
    case fa.WGSplatFragLayout():
      return True
    case fa.WGStridedFragLayout():
      return False
    case fa.TiledLayout():
      is_warp_replicated = any(isinstance(d, fa.Replicated) for d in layout.warp_dims)
      is_lane_replicated = any(isinstance(d, fa.Replicated) for d in layout.lane_dims)
      return is_warp_replicated or is_lane_replicated
    case _ as unreachable:
      return assert_never(unreachable)  # pytype: disable=wrong-arg-types


_tile_transform_attr_pattern = re.compile(
    r"^#mosaic_gpu.tile<[^>]+>$"
)


def is_tile_transform(attr: ir.Attribute) -> bool:
  return bool(_tile_transform_attr_pattern.search(str(attr)))


_transpose_transform_attr_pattern = re.compile(
    r"^#mosaic_gpu.transpose<[^>]+>$"
)


def is_transpose_transform(attr: ir.Attribute) -> bool:
  return bool(_transpose_transform_attr_pattern.search(str(attr)))


_swizzle_transform_attr_pattern = re.compile(
    r"^#mosaic_gpu.swizzle<[^>]+>$"
)

def is_swizzle_transform(attr: ir.Attribute) -> bool:
  return bool(_swizzle_transform_attr_pattern.search(str(attr)))


def to_transform_attr(
    transform: launch_context.MemRefTransform | mgpu.SwizzlingMode,
) -> ir.Attribute:
  if isinstance(transform, launch_context.TileTransform):
    return mgpu.TileTransformAttr.get(transform.tiling)
  elif isinstance(transform, launch_context.TransposeTransform):
    return mgpu.TransposeTransformAttr.get(transform.permutation)
  elif isinstance(transform, mgpu.SwizzlingMode):
    return mgpu.SwizzleTransformAttr.get(transform)
  else:
    raise NotImplementedError(f"Unsupported transform {transform}")


def from_transform_attr(
    transform: ir.Attribute,
) -> launch_context.MemRefTransform | mgpu.SwizzlingMode:
  if is_tile_transform(transform):
    return launch_context.TileTransform(
        mgpu.TileTransformAttr(transform).tiling
    )
  elif is_transpose_transform(transform):
    return launch_context.TransposeTransform(
        mgpu.TransposeTransformAttr(transform).permutation
    )
  elif is_swizzle_transform(transform):
    return mgpu.SwizzlingMode(mgpu.SwizzleTransformAttr(transform).swizzle)
  else:
    raise NotImplementedError(f"Unsupported transform {transform}")
